"""
Custom exceptions for the Disagreement library.
"""

from typing import Optional, Any


class DisagreementException(Exception):
    """Base exception class for all errors raised by this library."""

    pass


class HTTPException(DisagreementException):
    """Exception raised for HTTP-related errors.

    Attributes:
        response: The aiohttp response object, if available.
        status: The HTTP status code.
        text: The response text, if available.
        error_code: Discord specific error code, if available.
    """

    def __init__(
        self, response=None, message=None, *, status=None, text=None, error_code=None
    ):
        self.response = response
        self.status = status or (response.status if response else None)
        self.text = text or (
            response.text if response else None
        )  # Or await response.text() if in async context
        self.error_code = error_code

        full_message = f"HTTP {self.status}"
        if message:
            full_message += f": {message}"
        elif self.text:
            full_message += f": {self.text}"
        if self.error_code:
            full_message += f" (Discord Error Code: {self.error_code})"

        super().__init__(full_message)


class GatewayException(DisagreementException):
    """Exception raised for errors related to the Discord Gateway connection or protocol."""

    pass


class AuthenticationError(DisagreementException):
    """Exception raised for authentication failures (e.g., invalid token)."""

    pass


class RateLimitError(HTTPException):
    """
    Exception raised when a rate limit is encountered.

    Attributes:
        retry_after (float): The number of seconds to wait before retrying.
        is_global (bool): Whether this is a global rate limit.
    """

    def __init__(
        self, response, message=None, *, retry_after: float, is_global: bool = False
    ):
        self.retry_after = retry_after
        self.is_global = is_global
        super().__init__(
            response,
            message
            or f"Rate limited. Retry after: {retry_after}s. Global: {is_global}",
        )


# Specific HTTP error exceptions


class NotFound(HTTPException):
    """Raised for 404 Not Found errors."""

    pass


class Forbidden(HTTPException):
    """Raised for 403 Forbidden errors."""

    pass


class GeneralError(HTTPException):
    """General error (such as a malformed request body, amongst other things) (Code: 0)"""

    pass


class UnknownAccount(HTTPException):
    """Unknown account (Code: 10001)"""

    pass


class UnknownApplication(HTTPException):
    """Unknown application (Code: 10002)"""

    pass


class UnknownChannel(HTTPException):
    """Unknown channel (Code: 10003)"""

    pass


class UnknownGuild(HTTPException):
    """Unknown guild (Code: 10004)"""

    pass


class UnknownIntegration(HTTPException):
    """Unknown integration (Code: 10005)"""

    pass


class UnknownInvite(HTTPException):
    """Unknown invite (Code: 10006)"""

    pass


class UnknownMember(HTTPException):
    """Unknown member (Code: 10007)"""

    pass


class UnknownMessage(HTTPException):
    """Unknown message (Code: 10008)"""

    pass


class UnknownPermissionOverwrite(HTTPException):
    """Unknown permission overwrite (Code: 10009)"""

    pass


class UnknownProvider(HTTPException):
    """Unknown provider (Code: 10010)"""

    pass


class UnknownRole(HTTPException):
    """Unknown role (Code: 10011)"""

    pass


class UnknownToken(HTTPException):
    """Unknown token (Code: 10012)"""

    pass


class UnknownUser(HTTPException):
    """Unknown user (Code: 10013)"""

    pass


class UnknownEmoji(HTTPException):
    """Unknown emoji (Code: 10014)"""

    pass


class UnknownWebhook(HTTPException):
    """Unknown webhook (Code: 10015)"""

    pass


class UnknownWebhookService(HTTPException):
    """Unknown webhook service (Code: 10016)"""

    pass


class UnknownSession(HTTPException):
    """Unknown session (Code: 10020)"""

    pass


class UnknownAsset(HTTPException):
    """Unknown Asset (Code: 10021)"""

    pass


class UnknownBan(HTTPException):
    """Unknown ban (Code: 10026)"""

    pass


class UnknownSKU(HTTPException):
    """Unknown SKU (Code: 10027)"""

    pass


class UnknownStoreListing(HTTPException):
    """Unknown Store Listing (Code: 10028)"""

    pass


class UnknownEntitlement(HTTPException):
    """Unknown entitlement (Code: 10029)"""

    pass


class UnknownBuild(HTTPException):
    """Unknown build (Code: 10030)"""

    pass


class UnknownLobby(HTTPException):
    """Unknown lobby (Code: 10031)"""

    pass


class UnknownBranch(HTTPException):
    """Unknown branch (Code: 10032)"""

    pass


class UnknownStoreDirectoryLayout(HTTPException):
    """Unknown store directory layout (Code: 10033)"""

    pass


class UnknownRedistributable(HTTPException):
    """Unknown redistributable (Code: 10036)"""

    pass


class UnknownGiftCode(HTTPException):
    """Unknown gift code (Code: 10038)"""

    pass


class UnknownStream(HTTPException):
    """Unknown stream (Code: 10049)"""

    pass


class UnknownPremiumServerSubscribeCooldown(HTTPException):
    """Unknown premium server subscribe cooldown (Code: 10050)"""

    pass


class UnknownGuildTemplate(HTTPException):
    """Unknown guild template (Code: 10057)"""

    pass


class UnknownDiscoverableServerCategory(HTTPException):
    """Unknown discoverable server category (Code: 10059)"""

    pass


class UnknownSticker(HTTPException):
    """Unknown sticker (Code: 10060)"""

    pass


class UnknownStickerPack(HTTPException):
    """Unknown sticker pack (Code: 10061)"""

    pass


class UnknownInteraction(HTTPException):
    """Unknown interaction (Code: 10062)"""

    pass


class UnknownApplicationCommand(HTTPException):
    """Unknown application command (Code: 10063)"""

    pass


class UnknownVoiceState(HTTPException):
    """Unknown voice state (Code: 10065)"""

    pass


class UnknownApplicationCommandPermissions(HTTPException):
    """Unknown application command permissions (Code: 10066)"""

    pass


class UnknownStageInstance(HTTPException):
    """Unknown Stage Instance (Code: 10067)"""

    pass


class UnknownGuildMemberVerificationForm(HTTPException):
    """Unknown Guild Member Verification Form (Code: 10068)"""

    pass


class UnknownGuildWelcomeScreen(HTTPException):
    """Unknown Guild Welcome Screen (Code: 10069)"""

    pass


class UnknownGuildScheduledEvent(HTTPException):
    """Unknown Guild Scheduled Event (Code: 10070)"""

    pass


class UnknownGuildScheduledEventUser(HTTPException):
    """Unknown Guild Scheduled Event User (Code: 10071)"""

    pass


class UnknownTag(HTTPException):
    """Unknown Tag (Code: 10087)"""

    pass


class UnknownSound(HTTPException):
    """Unknown sound (Code: 10097)"""

    pass


class BotsCannotUseThisEndpoint(HTTPException):
    """Bots cannot use this endpoint (Code: 20001)"""

    pass


class OnlyBotsCanUseThisEndpoint(HTTPException):
    """Only bots can use this endpoint (Code: 20002)"""

    pass


class ExplicitContentCannotBeSentToTheDesiredRecipients(HTTPException):
    """Explicit content cannot be sent to the desired recipient(s) (Code: 20009)"""

    pass


class NotAuthorizedToPerformThisActionOnThisApplication(HTTPException):
    """You are not authorized to perform this action on this application (Code: 20012)"""

    pass


class ActionCannotBePerformedDueToSlowmodeRateLimit(HTTPException):
    """This action cannot be performed due to slowmode rate limit (Code: 20016)"""

    pass


class OnlyTheOwnerOfThisAccountCanPerformThisAction(HTTPException):
    """Only the owner of this account can perform this action (Code: 20018)"""

    pass


class MessageCannotBeEditedDueToAnnouncementRateLimits(HTTPException):
    """This message cannot be edited due to announcement rate limits (Code: 20022)"""

    pass


class UnderMinimumAge(HTTPException):
    """Under minimum age (Code: 20024)"""

    pass


class ChannelHitWriteRateLimit(HTTPException):
    """The channel you are writing has hit the write rate limit (Code: 20028)"""

    pass


class ServerHitWriteRateLimit(HTTPException):
    """The write action you are performing on the server has hit the write rate limit (Code: 20029)"""

    pass


class DisallowedWordsInStageTopicOrNames(HTTPException):
    """Your Stage topic, server name, server description, or channel names contain words that are not allowed (Code: 20031)"""

    pass


class GuildPremiumSubscriptionLevelTooLow(HTTPException):
    """Guild premium subscription level too low (Code: 20035)"""

    pass


class MaximumNumberOfGuildsReached(HTTPException):
    """Maximum number of guilds reached (100) (Code: 30001)"""

    pass


class MaximumNumberOfFriendsReached(HTTPException):
    """Maximum number of friends reached (1000) (Code: 30002)"""

    pass


class MaximumNumberOfPinsReached(HTTPException):
    """Maximum number of pins reached for the channel (50) (Code: 30003)"""

    pass


class MaximumNumberOfRecipientsReached(HTTPException):
    """Maximum number of recipients reached (10) (Code: 30004)"""

    pass


class MaximumNumberOfGuildRolesReached(HTTPException):
    """Maximum number of guild roles reached (250) (Code: 30005)"""

    pass


class MaximumNumberOfWebhooksReached(HTTPException):
    """Maximum number of webhooks reached (15) (Code: 30007)"""

    pass


class MaximumNumberOfEmojisReached(HTTPException):
    """Maximum number of emojis reached (Code: 30008)"""

    pass


class MaximumNumberOfReactionsReached(HTTPException):
    """Maximum number of reactions reached (20) (Code: 30010)"""

    pass


class MaximumNumberOfGroupDMsReached(HTTPException):
    """Maximum number of group DMs reached (10) (Code: 30011)"""

    pass


class MaximumNumberOfGuildChannelsReached(HTTPException):
    """Maximum number of guild channels reached (500) (Code: 30013)"""

    pass


class MaximumNumberOfAttachmentsInAMessageReached(HTTPException):
    """Maximum number of attachments in a message reached (10) (Code: 30015)"""

    pass


class MaximumNumberOfInvitesReached(HTTPException):
    """Maximum number of invites reached (1000) (Code: 30016)"""

    pass


class MaximumNumberOfAnimatedEmojisReached(HTTPException):
    """Maximum number of animated emojis reached (Code: 30018)"""

    pass


class MaximumNumberOfServerMembersReached(HTTPException):
    """Maximum number of server members reached (Code: 30019)"""

    pass


class MaximumNumberOfServerCategoriesReached(HTTPException):
    """Maximum number of server categories has been reached (5) (Code: 30030)"""

    pass


class GuildAlreadyHasATemplate(HTTPException):
    """Guild already has a template (Code: 30031)"""

    pass


class MaximumNumberOfApplicationCommandsReached(HTTPException):
    """Maximum number of application commands reached (Code: 30032)"""

    pass


class MaximumNumberOfThreadParticipantsReached(HTTPException):
    """Maximum number of thread participants has been reached (1000) (Code: 30033)"""

    pass


class MaximumNumberOfDailyApplicationCommandCreatesReached(HTTPException):
    """Maximum number of daily application command creates has been reached (200) (Code: 30034)"""

    pass


class MaximumNumberOfBansForNonGuildMembersExceeded(HTTPException):
    """Maximum number of bans for non-guild members have been exceeded (Code: 30035)"""

    pass


class MaximumNumberOfBansFetchesReached(HTTPException):
    """Maximum number of bans fetches has been reached (Code: 30037)"""

    pass


class MaximumNumberOfUncompletedGuildScheduledEventsReached(HTTPException):
    """Maximum number of uncompleted guild scheduled events reached (100) (Code: 30038)"""

    pass


class MaximumNumberOfStickersReached(HTTPException):
    """Maximum number of stickers reached (Code: 30039)"""

    pass


class MaximumNumberOfPruneRequestsReached(HTTPException):
    """Maximum number of prune requests has been reached. Try again later (Code: 30040)"""

    pass


class MaximumNumberOfGuildWidgetSettingsUpdatesReached(HTTPException):
    """Maximum number of guild widget settings updates has been reached. Try again later (Code: 30042)"""

    pass


class MaximumNumberOfSoundboardSoundsReached(HTTPException):
    """Maximum number of soundboard sounds reached (Code: 30045)"""

    pass


class MaximumNumberOfEditsToMessagesOlderThan1HourReached(HTTPException):
    """Maximum number of edits to messages older than 1 hour reached. Try again later (Code: 30046)"""

    pass


class MaximumNumberOfPinnedThreadsInAForumChannelReached(HTTPException):
    """Maximum number of pinned threads in a forum channel has been reached (Code: 30047)"""

    pass


class MaximumNumberOfTagsInAForumChannelReached(HTTPException):
    """Maximum number of tags in a forum channel has been reached (Code: 30048)"""

    pass


class BitrateIsTooHighForChannelOfThisType(HTTPException):
    """Bitrate is too high for channel of this type (Code: 30052)"""

    pass


class MaximumNumberOfPremiumEmojisReached(HTTPException):
    """Maximum number of premium emojis reached (25) (Code: 30056)"""

    pass


class MaximumNumberOfWebhooksPerGuildReached(HTTPException):
    """Maximum number of webhooks per guild reached (1000) (Code: 30058)"""

    pass


class MaximumNumberOfChannelPermissionOverwritesReached(HTTPException):
    """Maximum number of channel permission overwrites reached (1000) (Code: 30061)"""

    pass


class TheChannelsForThisGuildAreTooLarge(HTTPException):
    """The channels for this guild are too large (Code: 30061)"""

    pass


class Unauthorized(HTTPException):
    """Unauthorized. Provide a valid token and try again (Code: 40001)"""

    pass


class YouNeedToVerifyYourAccount(HTTPException):
    """You need to verify your account in order to perform this action (Code: 40002)"""

    pass


class YouAreOpeningDirectMessagesTooFast(HTTPException):
    """You are opening direct messages too fast (Code: 40003)"""

    pass


class SendMessagesHasBeenTemporarilyDisabled(HTTPException):
    """Send messages has been temporarily disabled (Code: 40004)"""

    pass


class RequestEntityTooLarge(HTTPException):
    """Request entity too large. Try sending something smaller in size (Code: 40005)"""

    pass


class ThisFeatureHasBeenTemporarilyDisabledServerSide(HTTPException):
    """This feature has been temporarily disabled server-side (Code: 40006)"""

    pass


class TheUserIsBannedFromThisGuild(HTTPException):
    """The user is banned from this guild (Code: 40007)"""

    pass


class ConnectionHasBeenRevoked(HTTPException):
    """Connection has been revoked (Code: 40012)"""

    pass


class OnlyConsumableSKUsCanBeConsumed(HTTPException):
    """Only consumable SKUs can be consumed (Code: 40018)"""

    pass


class YouCanOnlyDeleteSandboxEntitlements(HTTPException):
    """You can only delete sandbox entitlements. (Code: 40019)"""

    pass


class TargetUserIsNotConnectedToVoice(HTTPException):
    """Target user is not connected to voice (Code: 40032)"""

    pass


class ThisMessageHasAlreadyBeenCrossposted(HTTPException):
    """This message has already been crossposted (Code: 40033)"""

    pass


class AnApplicationCommandWithThatNameAlreadyExists(HTTPException):
    """An application command with that name already exists (Code: 40041)"""

    pass


class ApplicationInteractionFailedToSend(HTTPException):
    """Application interaction failed to send (Code: 40043)"""

    pass


class CannotSendAMessageInAForumChannel(HTTPException):
    """Cannot send a message in a forum channel (Code: 40058)"""

    pass


class InteractionHasAlreadyBeenAcknowledged(HTTPException):
    """Interaction has already been acknowledged (Code: 40060)"""

    pass


class TagNamesMustBeUnique(HTTPException):
    """Tag names must be unique (Code: 40061)"""

    pass


class ServiceResourceIsBeingRateLimited(HTTPException):
    """Service resource is being rate limited (Code: 40062)"""

    pass


class ThereAreNoTagsAvailableThatCanBeSetByNonModerators(HTTPException):
    """There are no tags available that can be set by non-moderators (Code: 40066)"""

    pass


class ATagIsRequiredToCreateAForumPostInThisChannel(HTTPException):
    """A tag is required to create a forum post in this channel (Code: 40067)"""

    pass


class AnEntitlementHasAlreadyBeenGrantedForThisResource(HTTPException):
    """An entitlement has already been granted for this resource (Code: 40074)"""

    pass


class ThisInteractionHasHitTheMaximumNumberOfFollowUpMessages(HTTPException):
    """This interaction has hit the maximum number of follow up messages (Code: 40094)"""

    pass


class CloudflareIsBlockingYourRequest(HTTPException):
    """Cloudflare is blocking your request. This can often be resolved by setting a proper User Agent (Code: 40333)"""

    pass


class MissingAccess(HTTPException):
    """Missing access (Code: 50001)"""

    pass


class InvalidAccountType(HTTPException):
    """Invalid account type (Code: 50002)"""

    pass


class CannotExecuteActionOnADMChannel(HTTPException):
    """Cannot execute action on a DM channel (Code: 50003)"""

    pass


class GuildWidgetDisabled(HTTPException):
    """Guild widget disabled (Code: 50004)"""

    pass


class CannotEditAMessageAuthoredByAnotherUser(HTTPException):
    """Cannot edit a message authored by another user (Code: 50005)"""

    pass


class CannotSendAnEmptyMessage(HTTPException):
    """Cannot send an empty message (Code: 50006)"""

    pass


class CannotSendMessagesToThisUser(HTTPException):
    """Cannot send messages to this user (Code: 50007)"""

    pass


class CannotSendMessagesInANonTextChannel(HTTPException):
    """Cannot send messages in a non-text channel (Code: 50008)"""

    pass


class ChannelVerificationLevelIsTooHighForYouToGainAccess(HTTPException):
    """Channel verification level is too high for you to gain access (Code: 50009)"""

    pass


class OAuth2ApplicationDoesNotHaveABot(HTTPException):
    """OAuth2 application does not have a bot (Code: 50010)"""

    pass


class OAuth2ApplicationLimitReached(HTTPException):
    """OAuth2 application limit reached (Code: 50011)"""

    pass


class InvalidOAuth2State(HTTPException):
    """Invalid OAuth2 state (Code: 50012)"""

    pass


class YouLackPermissionsToPerformThatAction(HTTPException):
    """You lack permissions to perform that action (Code: 50013)"""

    pass


class InvalidAuthenticationTokenProvided(HTTPException):
    """Invalid authentication token provided (Code: 50014)"""

    pass


class NoteWasTooLong(HTTPException):
    """Note was too long (Code: 50015)"""

    pass


class ProvidedTooFewOrTooManyMessagesToDelete(HTTPException):
    """Provided too few or too many messages to delete. Must provide at least 2 and fewer than 100 messages to delete (Code: 50016)"""

    pass


class InvalidMFALevel(HTTPException):
    """Invalid MFA Level (Code: 50017)"""

    pass


class AMessageCanOnlyBePinnedToTheChannelItWasSentIn(HTTPException):
    """A message can only be pinned to the channel it was sent in (Code: 50019)"""

    pass


class InviteCodeWasEitherInvalidOrTaken(HTTPException):
    """Invite code was either invalid or taken (Code: 50020)"""

    pass


class CannotExecuteActionOnASystemMessage(HTTPException):
    """Cannot execute action on a system message (Code: 50021)"""

    pass


class CannotExecuteActionOnThisChannelType(HTTPException):
    """Cannot execute action on this channel type (Code: 50024)"""

    pass


class InvalidOAuth2AccessTokenProvided(HTTPException):
    """Invalid OAuth2 access token provided (Code: 50025)"""

    pass


class MissingRequiredOAuth2Scope(HTTPException):
    """Missing required OAuth2 scope (Code: 50026)"""

    pass


class InvalidWebhookTokenProvided(HTTPException):
    """Invalid webhook token provided (Code: 50027)"""

    pass


class InvalidRole(HTTPException):
    """Invalid role (Code: 50028)"""

    pass


class InvalidRecipients(HTTPException):
    """Invalid Recipient(s) (Code: 50033)"""

    pass


class AMessageProvidedWasTooOldToBulkDelete(HTTPException):
    """A message provided was too old to bulk delete (Code: 50034)"""

    pass


class InvalidFormBody(HTTPException):
    """Invalid form body (returned for both application/json and multipart/form-data bodies), or invalid Content-Type provided (Code: 50035)"""

    pass


class AnInviteWasAcceptedToAGuildTheApplicationBotIsNotIn(HTTPException):
    """An invite was accepted to a guild the application's bot is not in (Code: 50036)"""

    pass


class InvalidActivityAction(HTTPException):
    """Invalid Activity Action (Code: 50039)"""

    pass


class InvalidAPIVersionProvided(HTTPException):
    """Invalid API version provided (Code: 50041)"""

    pass


class FileUploadedExceedsTheMaximumSize(HTTPException):
    """File uploaded exceeds the maximum size (Code: 50045)"""

    pass


class InvalidFileUploaded(HTTPException):
    """Invalid file uploaded (Code: 50046)"""

    pass


class CannotSelfRedeemThisGift(HTTPException):
    """Cannot self-redeem this gift (Code: 50054)"""

    pass


class InvalidGuild(HTTPException):
    """Invalid Guild (Code: 50055)"""

    pass


class InvalidSKU(HTTPException):
    """Invalid SKU (Code: 50057)"""

    pass


class InvalidRequestOrigin(HTTPException):
    """Invalid request origin (Code: 50067)"""

    pass


class InvalidMessageType(HTTPException):
    """Invalid message type (Code: 50068)"""

    pass


class PaymentSourceRequiredToRedeemGift(HTTPException):
    """Payment source required to redeem gift (Code: 50070)"""

    pass


class CannotModifyASystemWebhook(HTTPException):
    """Cannot modify a system webhook (Code: 50073)"""

    pass


class CannotDeleteAChannelRequiredForCommunityGuilds(HTTPException):
    """Cannot delete a channel required for Community guilds (Code: 50074)"""

    pass


class CannotEditStickersWithinAMessage(HTTPException):
    """Cannot edit stickers within a message (Code: 50080)"""

    pass


class InvalidStickerSent(HTTPException):
    """Invalid sticker sent (Code: 50081)"""

    pass


class TriedToPerformAnOperationOnAnArchivedThread(HTTPException):
    """Tried to perform an operation on an archived thread, such as editing a message or adding a user to the thread (Code: 50083)"""

    pass


class InvalidThreadNotificationSettings(HTTPException):
    """Invalid thread notification settings (Code: 50085)"""

    pass


class BeforeValueIsEarlierThanTheThreadCreationDate(HTTPException):
    """before value is earlier than the thread creation date (Code: 50086)"""

    pass


class CommunityServerChannelsMustBeTextChannels(HTTPException):
    """Community server channels must be text channels (Code: 50086)"""

    pass


class TheEntityTypeOfTheEventIsDifferentFromTheEntityYouAreTryingToStartTheEventFor(
    HTTPException
):
    """The entity type of the event is different from the entity you are trying to start the event for (Code: 50091)"""

    pass


class ThisServerIsNotAvailableInYourLocation(HTTPException):
    """This server is not available in your location (Code: 50095)"""

    pass


class ThisServerNeedsMonetizationEnabledInOrderToPerformThisAction(HTTPException):
    """This server needs monetization enabled in order to perform this action (Code: 50097)"""

    pass


class ThisServerNeedsMoreBoostsToPerformThisAction(HTTPException):
    """This server needs more boosts to perform this action (Code: 50101)"""

    pass


class TheRequestBodyContainsInvalidJSON(HTTPException):
    """The request body contains invalid JSON. (Code: 50109)"""

    pass


class TheProvidedFileIsInvalid(HTTPException):
    """The provided file is invalid. (Code: 50110)"""

    pass


class TheProvidedFileTypeIsInvalid(HTTPException):
    """The provided file type is invalid. (Code: 50123)"""

    pass


class TheProvidedFileDurationExceedsMaximumOf52Seconds(HTTPException):
    """The provided file duration exceeds maximum of 5.2 seconds. (Code: 50124)"""

    pass


class OwnerCannotBePendingMember(HTTPException):
    """Owner cannot be pending member (Code: 50131)"""

    pass


class OwnershipCannotBeTransferredToABotUser(HTTPException):
    """Ownership cannot be transferred to a bot user (Code: 50132)"""

    pass


class FailedToResizeAssetBelowTheMaximumSize(HTTPException):
    """Failed to resize asset below the maximum size: 262144 (Code: 50138)"""

    pass


class CannotMixSubscriptionAndNonSubscriptionRolesForAnEmoji(HTTPException):
    """Cannot mix subscription and non subscription roles for an emoji (Code: 50144)"""

    pass


class CannotConvertBetweenPremiumEmojiAndNormalEmoji(HTTPException):
    """Cannot convert between premium emoji and normal emoji (Code: 50145)"""

    pass


class UploadedFileNotFound(HTTPException):
    """Uploaded file not found. (Code: 50146)"""

    pass


class TheSpecifiedEmojiIsInvalid(HTTPException):
    """The specified emoji is invalid (Code: 50151)"""

    pass


class VoiceMessagesDoNotSupportAdditionalContent(HTTPException):
    """Voice messages do not support additional content. (Code: 50159)"""

    pass


class VoiceMessagesMustHaveASingleAudioAttachment(HTTPException):
    """Voice messages must have a single audio attachment. (Code: 50160)"""

    pass


class VoiceMessagesMustHaveSupportingMetadata(HTTPException):
    """Voice messages must have supporting metadata. (Code: 50161)"""

    pass


class VoiceMessagesCannotBeEdited(HTTPException):
    """Voice messages cannot be edited. (Code: 50162)"""

    pass


class CannotDeleteGuildSubscriptionIntegration(HTTPException):
    """Cannot delete guild subscription integration (Code: 50163)"""

    pass


class YouCannotSendVoiceMessagesInThisChannel(HTTPException):
    """You cannot send voice messages in this channel. (Code: 50173)"""

    pass


class TheUserAccountMustFirstBeVerified(HTTPException):
    """The user account must first be verified (Code: 50178)"""

    pass


class TheProvidedFileDoesNotHaveAValidDuration(HTTPException):
    """The provided file does not have a valid duration. (Code: 50192)"""

    pass


class YouDoNotHavePermissionToSendThisSticker(HTTPException):
    """You do not have permission to send this sticker. (Code: 50600)"""

    pass


class TwoFactorIsRequiredForThisOperation(HTTPException):
    """Two factor is required for this operation (Code: 60003)"""

    pass


class NoUsersWithDiscordTagExist(HTTPException):
    """No users with DiscordTag exist (Code: 80004)"""

    pass


class ReactionWasBlocked(HTTPException):
    """Reaction was blocked (Code: 90001)"""

    pass


class UserCannotUseBurstReactions(HTTPException):
    """User cannot use burst reactions (Code: 90002)"""

    pass


class ApplicationNotYetAvailable(HTTPException):
    """Application not yet available. Try again later (Code: 110001)"""

    pass


class APIResourceIsCurrentlyOverloaded(HTTPException):
    """API resource is currently overloaded. Try again a little later (Code: 130000)"""

    pass


class TheStageIsAlreadyOpen(HTTPException):
    """The Stage is already open (Code: 150006)"""

    pass


class CannotReplyWithoutPermissionToReadMessageHistory(HTTPException):
    """Cannot reply without permission to read message history (Code: 160002)"""

    pass


class AThreadHasAlreadyBeenCreatedForThisMessage(HTTPException):
    """A thread has already been created for this message (Code: 160004)"""

    pass


class ThreadIsLocked(HTTPException):
    """Thread is locked (Code: 160005)"""

    pass


class MaximumNumberOfActiveThreadsReached(HTTPException):
    """Maximum number of active threads reached (Code: 160006)"""

    pass


class MaximumNumberOfActiveAnnouncementThreadsReached(HTTPException):
    """Maximum number of active announcement threads reached (Code: 160007)"""

    pass


class InvalidJSONForUploadedLottieFile(HTTPException):
    """Invalid JSON for uploaded Lottie file (Code: 170001)"""

    pass


class UploadedLottiesCannotContainRasterizedImages(HTTPException):
    """Uploaded Lotties cannot contain rasterized images such as PNG or JPEG (Code: 170002)"""

    pass


class StickerMaximumFramerateExceeded(HTTPException):
    """Sticker maximum framerate exceeded (Code: 170003)"""

    pass


class StickerFrameCountExceedsMaximumOf1000Frames(HTTPException):
    """Sticker frame count exceeds maximum of 1000 frames (Code: 170004)"""

    pass


class LottieAnimationMaximumDimensionsExceeded(HTTPException):
    """Lottie animation maximum dimensions exceeded (Code: 170005)"""

    pass


class StickerFrameRateIsEitherTooSmallOrTooLarge(HTTPException):
    """Sticker frame rate is either too small or too large (Code: 170006)"""

    pass


class StickerAnimationDurationExceedsMaximumOf5Seconds(HTTPException):
    """Sticker animation duration exceeds maximum of 5 seconds (Code: 170007)"""

    pass


class CannotUpdateAFinishedEvent(HTTPException):
    """Cannot update a finished event (Code: 180000)"""

    pass


class FailedToCreateStageNeededForStageEvent(HTTPException):
    """Failed to create stage needed for stage event (Code: 180002)"""

    pass


class MessageWasBlockedByAutomaticModeration(HTTPException):
    """Message was blocked by automatic moderation (Code: 200000)"""

    pass


class TitleWasBlockedByAutomaticModeration(HTTPException):
    """Title was blocked by automatic moderation (Code: 200001)"""

    pass


class WebhooksPostedToForumChannelsMustHaveAThreadNameOrThreadId(HTTPException):
    """Webhooks posted to forum channels must have a thread_name or thread_id (Code: 220001)"""

    pass


class WebhooksPostedToForumChannelsCannotHaveBothAThreadNameAndThreadId(HTTPException):
    """Webhooks posted to forum channels cannot have both a thread_name and thread_id (Code: 220002)"""

    pass


class WebhooksCanOnlyCreateThreadsInForumChannels(HTTPException):
    """Webhooks can only create threads in forum channels (Code: 220003)"""

    pass


class WebhookServicesCannotBeUsedInForumChannels(HTTPException):
    """Webhook services cannot be used in forum channels (Code: 220004)"""

    pass


class MessageBlockedByHarmfulLinksFilter(HTTPException):
    """Message blocked by harmful links filter (Code: 240000)"""

    pass


class CannotEnableOnboardingRequirementsAreNotMet(HTTPException):
    """Cannot enable onboarding, requirements are not met (Code: 350000)"""

    pass


class CannotUpdateOnboardingWhileBelowRequirements(HTTPException):
    """Cannot update onboarding while below requirements (Code: 350001)"""

    pass


class FailedToBanUsers(HTTPException):
    """Failed to ban users (Code: 500000)"""

    pass


class PollVotingBlocked(HTTPException):
    """Poll voting blocked (Code: 520000)"""

    pass


class PollExpired(HTTPException):
    """Poll expired (Code: 520001)"""

    pass


class InvalidChannelTypeForPollCreation(HTTPException):
    """Invalid channel type for poll creation (Code: 520002)"""

    pass


class CannotEditAPollMessage(HTTPException):
    """Cannot edit a poll message (Code: 520003)"""

    pass


class CannotUseAnEmojiIncludedWithThePoll(HTTPException):
    """Cannot use an emoji included with the poll (Code: 520004)"""

    pass


class CannotExpireANonPollMessage(HTTPException):
    """Cannot expire a non-poll message (Code: 520006)"""

    pass


class AppCommandError(DisagreementException):
    """Base exception for application command related errors."""

    pass


class AppCommandOptionConversionError(AppCommandError):
    """Exception raised when an application command option fails to convert."""

    def __init__(
        self,
        message: str,
        option_name: Optional[str] = None,
        original_value: Any = None,
    ):
        self.option_name = option_name
        self.original_value = original_value
        full_message = message
        if option_name:
            full_message = f"Failed to convert option '{option_name}': {message}"
        if original_value is not None:
            full_message += f" (Original value: '{original_value}')"
        super().__init__(full_message)



import pytest
from types import SimpleNamespace
from unittest.mock import AsyncMock

from disagreement.http import HTTPClient
from disagreement.models import StageChannel, StageInstance
from disagreement.enums import ChannelType


@pytest.mark.asyncio
async def test_http_start_stage_instance_calls_request():
    http = HTTPClient(token="t")
    http.request = AsyncMock(return_value={"id": "1", "channel_id": "c", "topic": "t"})
    payload = {"channel_id": "c", "topic": "t", "privacy_level": 2}
    instance = await http.start_stage_instance(payload)
    http.request.assert_called_once_with(
        "POST",
        "/stage-instances",
        payload=payload,
        custom_headers=None,
    )
    assert isinstance(instance, StageInstance)


@pytest.mark.asyncio
async def test_http_end_stage_instance_calls_request():
    http = HTTPClient(token="t")
    http.request = AsyncMock(return_value=None)
    await http.end_stage_instance("c")
    http.request.assert_called_once_with(
        "DELETE", "/stage-instances/c", custom_headers=None
    )


@pytest.mark.asyncio
async def test_stage_channel_start_and_end():
    http = SimpleNamespace(
        start_stage_instance=AsyncMock(
            return_value=StageInstance({"id": "1", "channel_id": "c", "topic": "hi"})
        ),
        edit_stage_instance=AsyncMock(
            return_value=StageInstance({"id": "1", "channel_id": "c", "topic": "hi"})
        ),
        end_stage_instance=AsyncMock(),
    )
    client = type("Client", (), {})()
    client._http = http
    channel_data = {
        "id": "c",
        "type": ChannelType.GUILD_STAGE_VOICE.value,
        "guild_id": "g",
    }
    channel = StageChannel(channel_data, client)

    instance = await channel.start_stage_instance("hi")
    http.start_stage_instance.assert_awaited_once_with(
        {"channel_id": "c", "topic": "hi", "privacy_level": 2}, reason=None
    )
    assert isinstance(instance, StageInstance)
    assert instance._client is client

    await channel.end_stage_instance()
    http.end_stage_instance.assert_awaited_once_with("c", reason=None)



import pytest  # pylint: disable=E0401

from disagreement.models import Guild, Member, Role, TextChannel, PermissionOverwrite
from disagreement.enums import (
    ChannelType,
    VerificationLevel,
    MessageNotificationLevel,
    ExplicitContentFilterLevel,
    MFALevel,
    GuildNSFWLevel,
    PremiumTier,
    OverwriteType,
)
from disagreement.permissions import Permissions


from disagreement.client import Client


class DummyClient(Client):
    def __init__(self):
        super().__init__(token="test")


def _base_guild(client):
    data = {
        "id": "1",
        "name": "g",
        "owner_id": "1",
        "afk_timeout": 60,
        "verification_level": VerificationLevel.NONE.value,
        "default_message_notifications": MessageNotificationLevel.ALL_MESSAGES.value,
        "explicit_content_filter": ExplicitContentFilterLevel.DISABLED.value,
        "roles": [],
        "emojis": [],
        "features": [],
        "mfa_level": MFALevel.NONE.value,
        "system_channel_flags": 0,
        "premium_tier": PremiumTier.NONE.value,
        "nsfw_level": GuildNSFWLevel.DEFAULT.value,
    }
    guild = Guild(data, client_instance=client)
    client._guilds.set(guild.id, guild)
    return guild


def _member(guild, *roles):
    data = {
        "user": {"id": "10", "username": "u", "discriminator": "0001"},
        "joined_at": "t",
        "roles": [r.id for r in roles] or [guild.id],
    }
    member = Member(data, client_instance=None)
    member.guild_id = guild.id
    guild._members.set(member.id, member)
    return member


def _role(guild, rid, perms):
    role = Role(
        {
            "id": rid,
            "name": f"r{rid}",
            "color": 0,
            "hoist": False,
            "position": 0,
            "permissions": str(int(perms)),
            "managed": False,
            "mentionable": False,
        }
    )
    guild.roles.append(role)
    return role


def _channel(guild, client):
    data = {
        "id": "100",
        "type": ChannelType.GUILD_TEXT.value,
        "guild_id": guild.id,
        "permission_overwrites": [],
    }
    channel = TextChannel(data, client_instance=client)
    guild._channels.set(channel.id, channel)
    return channel


def test_permissions_for_base_roles():
    client = DummyClient()
    guild = _base_guild(client)
    everyone = _role(
        guild, guild.id, Permissions.VIEW_CHANNEL | Permissions.SEND_MESSAGES
    )
    mod = _role(guild, "2", Permissions.MANAGE_MESSAGES)
    member = _member(guild, everyone, mod)
    channel = _channel(guild, client)

    perms = channel.permissions_for(member)
    assert perms & Permissions.MANAGE_MESSAGES
    assert perms & Permissions.SEND_MESSAGES
    assert perms & Permissions.VIEW_CHANNEL


def test_permissions_for_with_overwrite():
    client = DummyClient()
    guild = _base_guild(client)
    everyone = _role(
        guild, guild.id, Permissions.VIEW_CHANNEL | Permissions.SEND_MESSAGES
    )
    mod = _role(guild, "2", Permissions.MANAGE_MESSAGES)
    member = _member(guild, everyone, mod)
    channel = _channel(guild, client)

    channel.permission_overwrites.append(
        PermissionOverwrite(
            {
                "id": mod.id,
                "type": OverwriteType.ROLE.value,
                "allow": "0",
                "deny": str(int(Permissions.MANAGE_MESSAGES)),
            }
        )
    )

    perms = channel.permissions_for(member)
    assert not perms & Permissions.MANAGE_MESSAGES
    assert perms & Permissions.SEND_MESSAGES
    assert perms & Permissions.VIEW_CHANNEL



import asyncio

import pytest  # pylint: disable=import-error

from disagreement.client import Client


@pytest.mark.asyncio
async def test_wait_for_resolves_on_event():
    client = Client(token="t")

    async def dispatch_event():
        await asyncio.sleep(0.05)
        data = {
            "id": "42",
            "channel_id": "c",
            "author": {"id": "1", "username": "u", "discriminator": "0001"},
            "content": "hello",
            "timestamp": "t",
        }
        await client._event_dispatcher.dispatch("MESSAGE_CREATE", data)

    asyncio.create_task(dispatch_event())
    message = await client.wait_for(
        "MESSAGE_CREATE", check=lambda m: m.id == "42", timeout=1
    )

    assert message.content == "hello"


@pytest.mark.asyncio
async def test_wait_for_timeout():
    client = Client(token="t")
    with pytest.raises(asyncio.TimeoutError):
        await client.wait_for("MESSAGE_CREATE", timeout=0.1)



import pytest
from types import SimpleNamespace
from unittest.mock import AsyncMock

from disagreement.client import Client
from disagreement.errors import DisagreementException
from disagreement.models import Message, User, Reaction


@pytest.mark.asyncio
async def test_create_reaction_calls_http():
    http = SimpleNamespace(create_reaction=AsyncMock())
    client = Client.__new__(Client)
    client._http = http
    client._closed = False

    message_data = {
        "id": "2",
        "channel_id": "1",
        "author": {"id": "3", "username": "u", "discriminator": "0001"},
        "content": "hi",
        "timestamp": "t",
    }
    message = Message(message_data, client_instance=client)

    await message.add_reaction("😀")

    http.create_reaction.assert_called_once_with("1", "2", "😀")


@pytest.mark.asyncio
async def test_create_reaction_closed():
    http = SimpleNamespace(create_reaction=AsyncMock())
    client = Client.__new__(Client)
    client._http = http
    client._closed = True

    message_data = {
        "id": "2",
        "channel_id": "1",
        "author": {"id": "3", "username": "u", "discriminator": "0001"},
        "content": "hi",
        "timestamp": "t",
    }
    message = Message(message_data, client_instance=client)

    with pytest.raises(DisagreementException):
        await message.add_reaction("😀")


@pytest.mark.asyncio
async def test_delete_reaction_calls_http():
    http = SimpleNamespace(delete_reaction=AsyncMock())
    client = Client.__new__(Client)
    client._http = http
    client._closed = False

    message_data = {
        "id": "2",
        "channel_id": "1",
        "author": {"id": "3", "username": "u", "discriminator": "0001"},
        "content": "hi",
        "timestamp": "t",
    }
    message = Message(message_data, client_instance=client)

    await message.remove_reaction("😀")

    http.delete_reaction.assert_called_once_with("1", "2", "😀")


@pytest.mark.asyncio
async def test_get_reactions_parses_users():
    users_payload = [{"id": "1", "username": "u", "discriminator": "0001"}]
    http = SimpleNamespace(get_reactions=AsyncMock(return_value=users_payload))
    client = Client(token="test")
    client._http = http

    users = await client.get_reactions("1", "2", "😀")

    http.get_reactions.assert_called_once_with("1", "2", "😀")
    assert isinstance(users[0], User)


@pytest.mark.asyncio
async def test_create_reaction_dispatches_event(monkeypatch):
    http = SimpleNamespace(create_reaction=AsyncMock())
    client = Client(token="t")
    client._http = http
    events = {}

    async def on_add(reaction):
        events["add"] = reaction

    client._event_dispatcher.register("MESSAGE_REACTION_ADD", on_add)

    message_data = {
        "id": "2",
        "channel_id": "1",
        "author": {"id": "3", "username": "u", "discriminator": "0001"},
        "content": "hi",
        "timestamp": "t",
    }
    message = Message(message_data, client_instance=client)

    await message.add_reaction("😀")

    assert isinstance(events.get("add"), Reaction)


@pytest.mark.asyncio
async def test_delete_reaction_dispatches_event(monkeypatch):
    http = SimpleNamespace(delete_reaction=AsyncMock())
    client = Client(token="t")
    client._http = http
    events = {}

    async def on_remove(reaction):
        events["remove"] = reaction

    client._event_dispatcher.register("MESSAGE_REACTION_REMOVE", on_remove)

    message_data = {
        "id": "2",
        "channel_id": "1",
        "author": {"id": "3", "username": "u", "discriminator": "0001"},
        "content": "hi",
        "timestamp": "t",
    }
    message = Message(message_data, client_instance=client)

    await message.remove_reaction("😀")

    assert isinstance(events.get("remove"), Reaction)



from __future__ import annotations

import operator
from typing import Any, Callable, ClassVar, Dict, Iterator, Tuple


class _MemberCacheFlagValue:
    flag: int

    def __init__(self, func: Callable[[Any], bool]):
        self.flag = getattr(func, "flag", 0)
        self.__doc__ = func.__doc__

    def __get__(self, instance: "MemberCacheFlags", owner: type) -> Any:
        if instance is None:
            return self
        return instance.value & self.flag != 0

    def __set__(self, instance: Any, value: bool) -> None:
        if value:
            instance.value |= self.flag
        else:
            instance.value &= ~self.flag

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} flag={self.flag}>"


def flag_value(flag: int) -> Callable[[Callable[[Any], bool]], _MemberCacheFlagValue]:
    def decorator(func: Callable[[Any], bool]) -> _MemberCacheFlagValue:
        setattr(func, "flag", flag)
        return _MemberCacheFlagValue(func)

    return decorator


class MemberCacheFlags:
    __slots__ = ("value",)

    VALID_FLAGS: ClassVar[Dict[str, int]] = {
        "joined": 1 << 0,
        "voice": 1 << 1,
        "online": 1 << 2,
    }
    DEFAULT_FLAGS: ClassVar[int] = 1 | 2 | 4
    ALL_FLAGS: ClassVar[int] = sum(VALID_FLAGS.values())

    def __init__(self, **kwargs: bool):
        self.value = self.DEFAULT_FLAGS
        for key, value in kwargs.items():
            if key not in self.VALID_FLAGS:
                raise TypeError(f"{key!r} is not a valid member cache flag.")
            setattr(self, key, value)

    @classmethod
    def _from_value(cls, value: int) -> MemberCacheFlags:
        self = cls.__new__(cls)
        self.value = value
        return self

    def __eq__(self, other: object) -> bool:
        return isinstance(other, MemberCacheFlags) and self.value == other.value

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash(self.value)

    def __repr__(self) -> str:
        return f"<MemberCacheFlags value={self.value}>"

    def __iter__(self) -> Iterator[Tuple[str, bool]]:
        for name in self.VALID_FLAGS:
            yield name, getattr(self, name)

    def __int__(self) -> int:
        return self.value

    def __index__(self) -> int:
        return self.value

    @classmethod
    def all(cls) -> MemberCacheFlags:
        """A factory method that creates a :class:`MemberCacheFlags` with all flags enabled."""
        return cls._from_value(cls.ALL_FLAGS)

    @classmethod
    def none(cls) -> MemberCacheFlags:
        """A factory method that creates a :class:`MemberCacheFlags` with all flags disabled."""
        return cls._from_value(0)

    @classmethod
    def only_joined(cls) -> MemberCacheFlags:
        """A factory method that creates a :class:`MemberCacheFlags` with only the `joined` flag enabled."""
        return cls._from_value(cls.VALID_FLAGS["joined"])

    @classmethod
    def only_voice(cls) -> MemberCacheFlags:
        """A factory method that creates a :class:`MemberCacheFlags` with only the `voice` flag enabled."""
        return cls._from_value(cls.VALID_FLAGS["voice"])

    @classmethod
    def only_online(cls) -> MemberCacheFlags:
        """A factory method that creates a :class:`MemberCacheFlags` with only the `online` flag enabled."""
        return cls._from_value(cls.VALID_FLAGS["online"])

    @flag_value(1 << 0)
    def joined(self) -> bool:
        """Whether to cache members that have just joined the guild."""
        return False

    @flag_value(1 << 1)
    def voice(self) -> bool:
        """Whether to cache members that are in a voice channel."""
        return False

    @flag_value(1 << 2)
    def online(self) -> bool:
        """Whether to cache members that are online."""
        return False



"""
Manages the WebSocket connection to the Discord Gateway.
"""

import asyncio
import logging
import traceback
import aiohttp
import json
import zlib
import time
import random
from typing import Optional, TYPE_CHECKING, Any, Dict

from .models import Activity

from .enums import GatewayOpcode, GatewayIntent
from .errors import GatewayException, DisagreementException, AuthenticationError
from .interactions import Interaction

if TYPE_CHECKING:
    from .client import Client  # For type hinting
    from .event_dispatcher import EventDispatcher
    from .http import HTTPClient
    from .interactions import Interaction  # Added for INTERACTION_CREATE

# ZLIB Decompression constants
ZLIB_SUFFIX = b"\x00\x00\xff\xff"
MAX_DECOMPRESSION_SIZE = 10 * 1024 * 1024  # 10 MiB, adjust as needed


logger = logging.getLogger(__name__)


class GatewayClient:
    """
    Handles the Discord Gateway WebSocket connection, heartbeating, and event dispatching.
    """

    def __init__(
        self,
        http_client: "HTTPClient",
        event_dispatcher: "EventDispatcher",
        token: str,
        intents: int,
        client_instance: "Client",  # Pass the main client instance
        verbose: bool = False,
        *,
        shard_id: Optional[int] = None,
        shard_count: Optional[int] = None,
        max_retries: int = 5,
        max_backoff: float = 60.0,
    ):
        self._http: "HTTPClient" = http_client
        self._dispatcher: "EventDispatcher" = event_dispatcher
        self._token: str = token
        self._intents: int = intents
        self._client_instance: "Client" = client_instance  # Store client instance
        self.verbose: bool = verbose
        self._shard_id: Optional[int] = shard_id
        self._shard_count: Optional[int] = shard_count
        self._max_retries: int = max_retries
        self._max_backoff: float = max_backoff

        self._ws: Optional[aiohttp.ClientWebSocketResponse] = None
        try:
            self._loop: asyncio.AbstractEventLoop = asyncio.get_running_loop()
        except RuntimeError:
            self._loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self._loop)
        self._heartbeat_interval: Optional[float] = None
        self._last_sequence: Optional[int] = None
        self._session_id: Optional[str] = None
        self._resume_gateway_url: Optional[str] = None

        self._keep_alive_task: Optional[asyncio.Task] = None
        self._receive_task: Optional[asyncio.Task] = None

        self._last_heartbeat_sent: Optional[float] = None
        self._last_heartbeat_ack: Optional[float] = None

        # For zlib decompression
        self._buffer = bytearray()
        self._inflator = zlib.decompressobj()

        self._member_chunk_requests: Dict[str, asyncio.Future] = {}

    async def _reconnect(self) -> None:
        """Attempts to reconnect using exponential backoff with jitter."""
        delay = 1.0
        for attempt in range(self._max_retries):
            try:
                await self.connect()
                return
            except Exception as e:  # noqa: BLE001
                if attempt >= self._max_retries - 1:
                    logger.error(
                        "Reconnect failed after %s attempts: %s", attempt + 1, e
                    )
                    raise
                jitter = random.uniform(0, delay)
                wait_time = min(delay + jitter, self._max_backoff)
                logger.warning(
                    "Reconnect attempt %s failed: %s. Retrying in %.2f seconds...",
                    attempt + 1,
                    e,
                    wait_time,
                )
                await asyncio.sleep(wait_time)
                delay = min(delay * 2, self._max_backoff)

    async def _decompress_message(
        self, message_bytes: bytes
    ) -> Optional[Dict[str, Any]]:
        """Decompresses a zlib-compressed message from the Gateway."""
        self._buffer.extend(message_bytes)

        if len(message_bytes) < 4 or message_bytes[-4:] != ZLIB_SUFFIX:
            # Message is not complete or not zlib compressed in the expected way
            return None
            # Or handle partial messages if Discord ever sends them fragmented like this,
            # but typically each binary message is a complete zlib stream.

        try:
            decompressed = self._inflator.decompress(self._buffer)
            self._buffer.clear()  # Reset buffer after successful decompression
            return json.loads(decompressed.decode("utf-8"))
        except zlib.error as e:
            logger.error("Zlib decompression error: %s", e)
            self._buffer.clear()  # Clear buffer on error
            self._inflator = zlib.decompressobj()  # Reset inflator
            return None
        except json.JSONDecodeError as e:
            logger.error("JSON decode error after decompression: %s", e)
            return None

    async def _send_json(self, payload: Dict[str, Any]):
        if self._ws and not self._ws.closed:
            if self.verbose:
                logger.debug("GATEWAY SEND: %s", payload)
            await self._ws.send_json(payload)
        else:
            logger.warning(
                "Gateway send attempted but WebSocket is closed or not available."
            )
            # raise GatewayException("WebSocket is not connected.")

    async def _heartbeat(self):
        """Sends a heartbeat to the Gateway."""
        self._last_heartbeat_sent = time.monotonic()
        payload = {"op": GatewayOpcode.HEARTBEAT, "d": self._last_sequence}
        await self._send_json(payload)

    async def _keep_alive(self):
        """Manages the heartbeating loop."""
        if self._heartbeat_interval is None:

            logger.error("Heartbeat interval not set. Cannot start keep_alive.")
            return

        try:
            while True:
                await self._heartbeat()
                await asyncio.sleep(
                    self._heartbeat_interval / 1000
                )  # Interval is in ms
        except asyncio.CancelledError:
            logger.debug("Keep_alive task cancelled.")
        except Exception as e:
            logger.error("Error in keep_alive loop: %s", e)
            # Potentially trigger a reconnect here or notify client
            await self._client_instance.close_gateway(code=1000)  # Generic close

    async def _identify(self):
        """Sends the IDENTIFY payload to the Gateway."""
        payload = {
            "op": GatewayOpcode.IDENTIFY,
            "d": {
                "token": self._token,
                "intents": self._intents,
                "properties": {
                    "$os": "python",  # Or platform.system()
                    "$browser": "disagreement",  # Library name
                    "$device": "disagreement",  # Library name
                },
                "compress": True,  # Request zlib compression
            },
        }
        if self._shard_id is not None and self._shard_count is not None:
            payload["d"]["shard"] = [self._shard_id, self._shard_count]
        await self._send_json(payload)
        logger.info("Sent IDENTIFY.")

    async def _resume(self):
        """Sends the RESUME payload to the Gateway."""
        if not self._session_id or self._last_sequence is None:
            logger.warning("Cannot RESUME: session_id or last_sequence is missing.")
            await self._identify()  # Fallback to identify
            return

        payload = {
            "op": GatewayOpcode.RESUME,
            "d": {
                "token": self._token,
                "session_id": self._session_id,
                "seq": self._last_sequence,
            },
        }
        await self._send_json(payload)
        logger.info(
            "Sent RESUME for session %s at sequence %s.",
            self._session_id,
            self._last_sequence,
        )

    async def update_presence(
        self,
        status: str,
        activity: Optional[Activity] = None,
        *,
        since: int = 0,
        afk: bool = False,
    ) -> None:
        """Sends the presence update payload to the Gateway."""
        payload = {
            "op": GatewayOpcode.PRESENCE_UPDATE,
            "d": {
                "since": since,
                "activities": [activity.to_dict()] if activity else [],
                "status": status,
                "afk": afk,
            },
        }
        await self._send_json(payload)

    async def request_guild_members(
        self,
        guild_id: str,
        query: str = "",
        limit: int = 0,
        presences: bool = False,
        user_ids: Optional[list[str]] = None,
        nonce: Optional[str] = None,
    ):
        """Sends the request guild members payload to the Gateway."""
        payload = {
            "op": GatewayOpcode.REQUEST_GUILD_MEMBERS,
            "d": {
                "guild_id": guild_id,
                "query": query,
                "limit": limit,
                "presences": presences,
            },
        }
        if user_ids:
            payload["d"]["user_ids"] = user_ids
        if nonce:
            payload["d"]["nonce"] = nonce

        await self._send_json(payload)

    async def _handle_dispatch(self, data: Dict[str, Any]):
        """Handles DISPATCH events (actual Discord events)."""
        event_name = data.get("t")
        sequence_num = data.get("s")
        raw_event_d_payload = data.get(
            "d"
        )  # This is the 'd' field from the gateway event

        if sequence_num is not None:
            self._last_sequence = sequence_num

        if event_name == "READY":  # Special handling for READY
            if not isinstance(raw_event_d_payload, dict):
                logger.error(
                    "READY event 'd' payload is not a dict or is missing: %s",
                    raw_event_d_payload,
                )
                # Consider raising an error or attempting a reconnect
                return
            self._session_id = raw_event_d_payload.get("session_id")
            self._resume_gateway_url = raw_event_d_payload.get("resume_gateway_url")

            app_id_str = "N/A"
            # Store application_id on the client instance
            if (
                "application" in raw_event_d_payload
                and isinstance(raw_event_d_payload["application"], dict)
                and "id" in raw_event_d_payload["application"]
            ):
                app_id_value = raw_event_d_payload["application"]["id"]
                self._client_instance.application_id = (
                    app_id_value  # Snowflake can be str or int
                )
                app_id_str = str(app_id_value)
            else:
                logger.warning(
                    "Could not find application ID in READY payload. App commands may not work."
                )

            # Parse and store the bot's own user object
            if "user" in raw_event_d_payload and isinstance(
                raw_event_d_payload["user"], dict
            ):
                try:
                    # Assuming Client has a parse_user method that takes user data dict
                    # and returns a User object, also caching it.
                    bot_user_obj = self._client_instance.parse_user(
                        raw_event_d_payload["user"]
                    )
                    self._client_instance.user = bot_user_obj
                    logger.info(
                        "Gateway READY. Bot User: %s#%s. Session ID: %s. App ID: %s. Resume URL: %s",
                        bot_user_obj.username,
                        bot_user_obj.discriminator,
                        self._session_id,
                        app_id_str,
                        self._resume_gateway_url,
                    )
                except Exception as e:
                    logger.error("Error parsing bot user from READY payload: %s", e)
                    logger.info(
                        "Gateway READY (user parse failed). Session ID: %s. App ID: %s. Resume URL: %s",
                        self._session_id,
                        app_id_str,
                        self._resume_gateway_url,
                    )
            else:
                logger.warning("Bot user object not found or invalid in READY payload.")
                logger.info(
                    "Gateway READY (no user). Session ID: %s. App ID: %s. Resume URL: %s",
                    self._session_id,
                    app_id_str,
                    self._resume_gateway_url,
                )

            await self._dispatcher.dispatch(event_name, raw_event_d_payload)
        elif event_name == "GUILD_MEMBERS_CHUNK":
            if isinstance(raw_event_d_payload, dict):
                nonce = raw_event_d_payload.get("nonce")
                if nonce and nonce in self._member_chunk_requests:
                    future = self._member_chunk_requests[nonce]
                    if not future.done():
                        # Append members to a temporary list stored on the future object
                        if not hasattr(future, "_members"):
                            future._members = []  # type: ignore
                        future._members.extend(raw_event_d_payload.get("members", []))  # type: ignore

                        # If this is the last chunk, resolve the future
                        if (
                            raw_event_d_payload.get("chunk_index")
                            == raw_event_d_payload.get("chunk_count", 1) - 1
                        ):
                            future.set_result(future._members)  # type: ignore
                            del self._member_chunk_requests[nonce]

        elif event_name == "INTERACTION_CREATE":

            if isinstance(raw_event_d_payload, dict):
                interaction = Interaction(
                    data=raw_event_d_payload, client_instance=self._client_instance
                )
                await self._dispatcher.dispatch(
                    "INTERACTION_CREATE", raw_event_d_payload
                )
                # Dispatch to a new client method that will then call AppCommandHandler
                if hasattr(self._client_instance, "process_interaction"):
                    asyncio.create_task(
                        self._client_instance.process_interaction(interaction)
                    )  # type: ignore
                else:
                    logger.warning(
                        "Client instance does not have process_interaction method for INTERACTION_CREATE."
                    )
            else:
                logger.error(
                    "INTERACTION_CREATE event 'd' payload is not a dict: %s",
                    raw_event_d_payload,
                )
        elif event_name == "RESUMED":
            logger.info("Gateway RESUMED successfully.")
            # RESUMED 'd' payload is often an empty object or debug info.
            # Ensure it's a dict for the dispatcher.
            event_data_to_dispatch = (
                raw_event_d_payload if isinstance(raw_event_d_payload, dict) else {}
            )
            await self._dispatcher.dispatch(event_name, event_data_to_dispatch)
            await self._dispatcher.dispatch(
                "SHARD_RESUME", {"shard_id": self._shard_id}
            )
        elif event_name:
            # For other events, ensure 'd' is a dict, or pass {} if 'd' is null/missing.
            # Models/parsers in EventDispatcher will need to handle potentially empty dicts.
            event_data_to_dispatch = (
                raw_event_d_payload if isinstance(raw_event_d_payload, dict) else {}
            )

            await self._dispatcher.dispatch(event_name, event_data_to_dispatch)
        else:
            logger.warning("Received dispatch with no event name: %s", data)

    async def _process_message(self, msg: aiohttp.WSMessage):
        """Processes a single message from the WebSocket."""
        if msg.type == aiohttp.WSMsgType.TEXT:
            try:
                data = json.loads(msg.data)
            except json.JSONDecodeError:
                logger.error("Failed to decode JSON from Gateway: %s", msg.data[:200])
                return
        elif msg.type == aiohttp.WSMsgType.BINARY:
            decompressed_data = await self._decompress_message(msg.data)
            if decompressed_data is None:
                logger.error(
                    "Failed to decompress or decode binary message from Gateway."
                )
                return
            data = decompressed_data
        elif msg.type == aiohttp.WSMsgType.ERROR:
            logger.error(
                "WebSocket error: %s",
                self._ws.exception() if self._ws else "Unknown WSError",
            )
            raise GatewayException(
                f"WebSocket error: {self._ws.exception() if self._ws else 'Unknown WSError'}"
            )
        elif msg.type == aiohttp.WSMsgType.CLOSED:
            close_code = (
                self._ws.close_code
                if self._ws and hasattr(self._ws, "close_code")
                else "N/A"
            )
            logger.warning(
                "WebSocket connection closed by server. Code: %s", close_code
            )
            # Raise an exception to signal the closure to the client's main run loop
            raise GatewayException(f"WebSocket closed by server. Code: {close_code}")
        else:
            logger.warning("Received unhandled WebSocket message type: %s", msg.type)
            return

        if self.verbose:
            logger.debug("GATEWAY RECV: %s", data)
        op = data.get("op")
        # 'd' payload (event_data) is handled specifically by each opcode handler below

        if op == GatewayOpcode.DISPATCH:
            await self._handle_dispatch(data)  # _handle_dispatch will extract 'd'
        elif op == GatewayOpcode.HEARTBEAT:  # Server requests a heartbeat
            await self._heartbeat()
        elif op == GatewayOpcode.RECONNECT:  # Server requests a reconnect
            logger.info(
                "Gateway requested RECONNECT. Closing and will attempt to reconnect."
            )
            await self.close(code=4000, reconnect=True)
        elif op == GatewayOpcode.INVALID_SESSION:
            # The 'd' payload for INVALID_SESSION is a boolean indicating resumability
            can_resume = data.get("d") is True
            logger.warning(
                "Gateway indicated INVALID_SESSION. Resumable: %s", can_resume
            )
            if not can_resume:
                self._session_id = None  # Clear session_id to force re-identify
                self._last_sequence = None
            # Close and reconnect. The connect logic will decide to resume or identify.
            await self.close(code=4000 if can_resume else 4009, reconnect=True)
        elif op == GatewayOpcode.HELLO:
            hello_d_payload = data.get("d")
            if (
                not isinstance(hello_d_payload, dict)
                or "heartbeat_interval" not in hello_d_payload
            ):
                logger.error(
                    "HELLO event 'd' payload is invalid or missing heartbeat_interval: %s",
                    hello_d_payload,
                )
                await self.close(code=1011)  # Internal error, malformed HELLO
                return
            self._heartbeat_interval = hello_d_payload["heartbeat_interval"]
            logger.info(
                "Gateway HELLO. Heartbeat interval: %sms.", self._heartbeat_interval
            )
            # Start heartbeating
            if self._keep_alive_task:
                self._keep_alive_task.cancel()
            self._keep_alive_task = self._loop.create_task(self._keep_alive())

            # Identify or Resume
            if self._session_id and self._resume_gateway_url:  # Check if we can resume
                logger.info("Attempting to RESUME session.")
                await self._resume()
            else:
                logger.info("Performing initial IDENTIFY.")
                await self._identify()
        elif op == GatewayOpcode.HEARTBEAT_ACK:
            self._last_heartbeat_ack = time.monotonic()
        else:
            logger.warning(
                "Received unhandled Gateway Opcode: %s with data: %s", op, data
            )

    async def _receive_loop(self):
        """Continuously receives and processes messages from the WebSocket."""
        if not self._ws or self._ws.closed:
            logger.warning(
                "Receive loop cannot start: WebSocket is not connected or closed."
            )
            return

        try:
            async for msg in self._ws:
                await self._process_message(msg)
        except asyncio.CancelledError:
            logger.debug("Receive_loop task cancelled.")
        except aiohttp.ClientConnectionError as e:
            logger.warning(
                "ClientConnectionError in receive_loop: %s. Attempting reconnect.", e
            )
            await self.close(code=1006, reconnect=True)  # Abnormal closure
        except Exception as e:
            logger.error("Unexpected error in receive_loop: %s", e)
            traceback.print_exc()
            await self.close(code=1011, reconnect=True)
        finally:
            logger.info("Receive_loop ended.")
            # If the loop ends unexpectedly (not due to explicit close),
            # the main client might want to try reconnecting.

    async def connect(self):
        """Connects to the Discord Gateway."""
        if self._ws and not self._ws.closed:
            logger.warning("Gateway already connected or connecting.")
            return

        gateway_url = (
            self._resume_gateway_url or (await self._http.get_gateway_bot())["url"]
        )
        if not gateway_url.endswith("?v=10&encoding=json&compress=zlib-stream"):
            gateway_url += "?v=10&encoding=json&compress=zlib-stream"

        logger.info("Connecting to Gateway: %s", gateway_url)
        try:
            await self._http._ensure_session()  # Ensure the HTTP client's session is active
            assert (
                self._http._session is not None
            ), "HTTPClient session not initialized after ensure_session"
            self._ws = await self._http._session.ws_connect(gateway_url, max_msg_size=0)
            logger.info("Gateway WebSocket connection established.")

            if self._receive_task:
                self._receive_task.cancel()
            self._receive_task = self._loop.create_task(self._receive_loop())

            await self._dispatcher.dispatch(
                "SHARD_CONNECT", {"shard_id": self._shard_id}
            )

        except aiohttp.ClientConnectorError as e:
            raise GatewayException(
                f"Failed to connect to Gateway (Connector Error): {e}"
            ) from e
        except aiohttp.WSServerHandshakeError as e:
            if e.status == 401:  # Unauthorized during handshake
                raise AuthenticationError(
                    f"Gateway handshake failed (401 Unauthorized): {e.message}. Check your bot token."
                ) from e
            raise GatewayException(
                f"Gateway handshake failed (Status: {e.status}): {e.message}"
            ) from e
        except Exception as e:  # Catch other potential errors during connection
            raise GatewayException(
                f"An unexpected error occurred during Gateway connection: {e}"
            ) from e

    async def close(self, code: int = 1000, *, reconnect: bool = False):
        """Closes the Gateway connection."""
        logger.info("Closing Gateway connection with code %s...", code)
        if self._keep_alive_task and not self._keep_alive_task.done():
            self._keep_alive_task.cancel()
            try:
                await self._keep_alive_task
            except asyncio.CancelledError:
                pass

        if self._receive_task and not self._receive_task.done():
            current = asyncio.current_task(loop=self._loop)
            self._receive_task.cancel()
            if self._receive_task is not current:
                try:
                    await self._receive_task
                except asyncio.CancelledError:
                    pass

        if self._ws and not self._ws.closed:
            await self._ws.close(code=code)
            logger.info("Gateway WebSocket closed.")

        self._ws = None
        # Do not reset session_id, last_sequence, or resume_gateway_url here
        # if the close code indicates a resumable disconnect (e.g. 4000-4009, or server-initiated RECONNECT)
        # The connect logic will decide whether to resume or re-identify.
        # However, if it's a non-resumable close (e.g. Invalid Session non-resumable), clear them.
        if code == 4009:  # Invalid session, not resumable
            logger.info("Clearing session state due to non-resumable invalid session.")
            self._session_id = None
            self._last_sequence = None
            self._resume_gateway_url = None  # This might be re-fetched anyway

        await self._dispatcher.dispatch(
            "SHARD_DISCONNECT", {"shard_id": self._shard_id}
        )

    @property
    def latency(self) -> Optional[float]:
        """Returns the latency between heartbeat and ACK in seconds."""
        if self._last_heartbeat_sent is None or self._last_heartbeat_ack is None:
            return None
        return self._last_heartbeat_ack - self._last_heartbeat_sent

    @property
    def latency_ms(self) -> Optional[float]:
        """Returns the latency between heartbeat and ACK in milliseconds."""
        if self._last_heartbeat_sent is None or self._last_heartbeat_ack is None:
            return None
        return (self._last_heartbeat_ack - self._last_heartbeat_sent) * 1000

    @property
    def last_heartbeat_sent(self) -> Optional[float]:
        return self._last_heartbeat_sent

    @property
    def last_heartbeat_ack(self) -> Optional[float]:
        return self._last_heartbeat_ack



# Example from the README file of the disagreement library
# This example demonstrates a simple bot that responds to the "!ping" command with "Pong!".

import asyncio
import os

import disagreement
from disagreement.ext import commands
from dotenv import load_dotenv

load_dotenv()


class Basics(commands.Cog):
    def __init__(self, client: disagreement.Client) -> None:
        super().__init__(client)

    @commands.command()
    async def ping(self, ctx: commands.CommandContext) -> None:
        await ctx.reply(f"Pong! Gateway Latency: {self.client.latency_ms} ms.")  # type: ignore (latency is None during static analysis)


token = os.getenv("DISCORD_BOT_TOKEN")
if not token:
    raise RuntimeError("DISCORD_BOT_TOKEN environment variable not set")

intents = (
    disagreement.GatewayIntent.default() | disagreement.GatewayIntent.MESSAGE_CONTENT
)
client = disagreement.Client(
    token=token, command_prefix="!", intents=intents, mention_replies=True
)


async def main() -> None:
    client.add_cog(Basics(client))
    await client.run()


if __name__ == "__main__":
    asyncio.run(main())



import pytest

from disagreement.ext.commands.converters import run_converters
from disagreement.ext.commands.core import CommandContext, Command
from disagreement.ext.commands.errors import BadArgument
from disagreement.models import Message, Member, Role, Guild
from disagreement.enums import (
    VerificationLevel,
    MessageNotificationLevel,
    ExplicitContentFilterLevel,
    MFALevel,
    GuildNSFWLevel,
    PremiumTier,
)


from disagreement.client import Client
from disagreement.cache import GuildCache


class DummyBot(Client):
    def __init__(self):
        super().__init__(token="test")
        self._guilds = GuildCache()

    def get_guild(self, guild_id):
        return self._guilds.get(guild_id)

    async def fetch_member(self, guild_id, member_id):
        guild = self._guilds.get(guild_id)
        return guild.get_member(member_id) if guild else None

    async def fetch_role(self, guild_id, role_id):
        guild = self._guilds.get(guild_id)
        return guild.get_role(role_id) if guild else None

    async def fetch_guild(self, guild_id):
        return self._guilds.get(guild_id)


@pytest.fixture()
def guild_objects():
    guild_data = {
        "id": "1",
        "name": "g",
        "owner_id": "2",
        "afk_timeout": 60,
        "verification_level": VerificationLevel.NONE.value,
        "default_message_notifications": MessageNotificationLevel.ALL_MESSAGES.value,
        "explicit_content_filter": ExplicitContentFilterLevel.DISABLED.value,
        "roles": [],
        "emojis": [],
        "features": [],
        "mfa_level": MFALevel.NONE.value,
        "system_channel_flags": 0,
        "premium_tier": PremiumTier.NONE.value,
        "nsfw_level": GuildNSFWLevel.DEFAULT.value,
    }
    bot = DummyBot()
    guild = Guild(guild_data, client_instance=bot)
    bot._guilds.set(guild.id, guild)

    member = Member(
        {
            "user": {"id": "3", "username": "m", "discriminator": "0001"},
            "joined_at": "t",
            "roles": [],
        },
        None,
    )
    member.guild_id = guild.id

    role = Role(
        {
            "id": "5",
            "name": "r",
            "color": 0,
            "hoist": False,
            "position": 0,
            "permissions": "0",
            "managed": False,
            "mentionable": True,
        }
    )

    guild._members.set(member.id, member)
    guild.roles.append(role)

    return guild, member, role


@pytest.fixture()
def command_context(guild_objects):
    guild, member, role = guild_objects
    bot = guild._client
    message_data = {
        "id": "10",
        "channel_id": "20",
        "guild_id": guild.id,
        "author": {"id": "2", "username": "u", "discriminator": "0001"},
        "content": "hi",
        "timestamp": "t",
    }
    msg = Message(message_data, client_instance=bot)

    async def dummy(ctx):
        pass

    cmd = Command(dummy)
    return CommandContext(
        message=msg, bot=bot, prefix="!", command=cmd, invoked_with="dummy"
    )


@pytest.mark.asyncio
async def test_member_converter(command_context, guild_objects):
    _, member, _ = guild_objects
    mention = f"<@!{member.id}>"
    result = await run_converters(command_context, Member, mention)
    assert result is member
    result = await run_converters(command_context, Member, member.id)
    assert result is member


@pytest.mark.asyncio
async def test_role_converter(command_context, guild_objects):
    _, _, role = guild_objects
    mention = f"<@&{role.id}>"
    result = await run_converters(command_context, Role, mention)
    assert result is role
    result = await run_converters(command_context, Role, role.id)
    assert result is role


@pytest.mark.asyncio
async def test_guild_converter(command_context, guild_objects):
    guild, _, _ = guild_objects
    result = await run_converters(command_context, Guild, guild.id)
    assert result is guild


@pytest.mark.asyncio
async def test_member_converter_no_guild():
    guild_data = {
        "id": "99",
        "name": "g",
        "owner_id": "2",
        "afk_timeout": 60,
        "verification_level": VerificationLevel.NONE.value,
        "default_message_notifications": MessageNotificationLevel.ALL_MESSAGES.value,
        "explicit_content_filter": ExplicitContentFilterLevel.DISABLED.value,
        "roles": [],
        "emojis": [],
        "features": [],
        "mfa_level": MFALevel.NONE.value,
        "system_channel_flags": 0,
        "premium_tier": PremiumTier.NONE.value,
        "nsfw_level": GuildNSFWLevel.DEFAULT.value,
    }
    bot = DummyBot()
    guild = Guild(guild_data, client_instance=bot)
    bot._guilds.set(guild.id, guild)
    message_data = {
        "id": "11",
        "channel_id": "20",
        "author": {"id": "2", "username": "u", "discriminator": "0001"},
        "content": "hi",
        "timestamp": "t",
    }
    msg = Message(message_data, client_instance=bot)

    async def dummy(ctx):
        pass

    ctx = CommandContext(
        message=msg, bot=bot, prefix="!", command=Command(dummy), invoked_with="dummy"
    )

    with pytest.raises(BadArgument):
        await run_converters(ctx, Member, "<@!1>")



import time

from disagreement.cache import Cache


def test_cache_store_and_get():
    cache = Cache()
    cache.set("a", 123)
    assert cache.get("a") == 123


def test_cache_ttl_expiry():
    cache = Cache(ttl=0.01)
    cache.set("b", 1)
    assert cache.get("b") == 1
    time.sleep(0.02)
    assert cache.get("b") is None


def test_cache_lru_eviction():
    cache = Cache(maxlen=2)
    cache.set("a", 1)
    cache.set("b", 2)
    assert cache.get("a") == 1
    cache.set("c", 3)
    assert cache.get("b") is None
    assert cache.get("a") == 1
    assert cache.get("c") == 3


def test_get_or_fetch_uses_cache():
    cache = Cache()
    cache.set("a", 1)

    def fetch():
        raise AssertionError("fetch should not be called")

    assert cache.get_or_fetch("a", fetch) == 1


def test_get_or_fetch_fetches_and_stores():
    cache = Cache()
    called = False

    def fetch():
        nonlocal called
        called = True
        return 2

    assert cache.get_or_fetch("b", fetch) == 2
    assert called
    assert cache.get("b") == 2


def test_get_or_fetch_fetches_expired_item():
    cache = Cache(ttl=0.01)
    cache.set("c", 1)
    time.sleep(0.02)
    called = False

    def fetch():
        nonlocal called
        called = True
        return 3

    assert cache.get_or_fetch("c", fetch) == 3
    assert called



import pytest
from types import SimpleNamespace
from unittest.mock import AsyncMock

from disagreement.http import HTTPClient
from disagreement.client import Client
from disagreement.models import GuildTemplate


@pytest.mark.asyncio
async def test_get_guild_templates_calls_request():
    http = HTTPClient(token="t")
    http.request = AsyncMock(return_value=[])
    await http.get_guild_templates("1")
    http.request.assert_called_once_with("GET", "/guilds/1/templates")


@pytest.mark.asyncio
async def test_client_fetch_templates_returns_models():
    http = SimpleNamespace(
        get_guild_templates=AsyncMock(
            return_value=[
                {
                    "code": "c",
                    "name": "n",
                    "usage_count": 0,
                    "creator_id": "u",
                    "created_at": "t",
                    "updated_at": "t",
                    "source_guild_id": "g",
                }
            ]
        )
    )
    client = Client.__new__(Client)
    client._http = http
    client._closed = False

    templates = await client.fetch_templates("g")

    http.get_guild_templates.assert_awaited_once_with("g")
    assert isinstance(templates[0], GuildTemplate)



# examples/reactions.py

"""
An example bot demonstrating reaction handling with the Disagreement library.

This bot will:
1. React to a specific command with a thumbs-up emoji.
2. Log when any reaction is added to a message in a server it's in.
3. Log when any reaction is removed from a message.

To run this bot:
1. Follow the setup steps in 'basic_bot.py' to set your DISCORD_BOT_TOKEN.
2. Ensure you have the GUILD_MESSAGE_REACTIONS intent enabled for your bot in the Discord Developer Portal.
3. Run this script: python examples/reactions.py
"""

import asyncio
import os
import sys
import logging
import traceback

# Add project root to path for local development
if os.path.join(os.getcwd(), "examples") == os.path.dirname(os.path.abspath(__file__)):
    sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

try:
    import disagreement
    from disagreement.ext import commands
    from disagreement.models import Reaction, User, Member
except ImportError:
    print(
        "Failed to import disagreement. Make sure it's installed or PYTHONPATH is set correctly."
    )
    sys.exit(1)

try:
    from dotenv import load_dotenv
except ImportError:  # pragma: no cover - example helper
    load_dotenv = None
    print("python-dotenv is not installed. Environment variables will not be loaded")

if load_dotenv:
    load_dotenv()

# --- Bot Configuration ---
BOT_TOKEN = os.environ.get("DISCORD_BOT_TOKEN")

# --- Intents Configuration ---
# We need GUILDS for server context, GUILD_MESSAGES to receive messages,
# and GUILD_MESSAGE_REACTIONS to listen for reaction events.
intents = (
    disagreement.GatewayIntent.GUILDS
    | disagreement.GatewayIntent.GUILD_MESSAGES
    | disagreement.GatewayIntent.GUILD_MESSAGE_REACTIONS
    | disagreement.GatewayIntent.MESSAGE_CONTENT  # For commands
)

# --- Initialize the Client ---
if not BOT_TOKEN:
    print("Error: The DISCORD_BOT_TOKEN environment variable is not set.")
    sys.exit(1)

client = disagreement.Client(token=BOT_TOKEN, intents=intents, command_prefix="!")


# --- Define a Cog for reaction-related commands ---
class ReactionCog(commands.Cog):
    def __init__(self, bot_client):
        super().__init__(bot_client)

    @commands.command(name="react")
    async def react_command(self, ctx: commands.CommandContext):
        """Reacts to the command message with a thumbs up."""
        try:
            # The emoji can be a standard Unicode emoji or a custom one in the format '<:name:id>'
            await ctx.message.add_reaction("👍")
            print(f"Reacted to command from {ctx.author.username}")
        except disagreement.HTTPException as e:
            print(f"Failed to add reaction: {e}")
            await ctx.reply(
                "I couldn't add the reaction. I might be missing permissions."
            )


# --- Event Handlers ---


@client.event
async def on_ready():
    """Called when the bot is ready and connected to Discord."""
    if client.user:
        print(f"Bot is ready! Logged in as {client.user.username}")
    else:
        print("Bot is ready, but client.user is missing!")
    print("------")
    print("Reaction example bot is operational.")


@client.on_event("MESSAGE_REACTION_ADD")
async def on_reaction_add(reaction: Reaction, user: User | Member):
    """Called when a message reaction is added."""
    # We can ignore reactions from the bot itself
    if client.user and user.id == client.user.id:
        return

    print(
        f"Reaction '{reaction.emoji}' added by {user.username} "
        f"to message ID {reaction.message_id} in channel ID {reaction.channel_id}"
    )
    # You can fetch the message if you need its content, but it's an extra API call.
    # try:
    #     channel = await client.fetch_channel(reaction.channel_id)
    #     if isinstance(channel, disagreement.TextChannel):
    #         message = await channel.fetch_message(reaction.message_id)
    #         print(f"  Message content: '{message.content}'")
    # except disagreement.errors.NotFound:
    #     print("  Could not fetch message (maybe it was deleted).")


@client.on_event("MESSAGE_REACTION_REMOVE")
async def on_reaction_remove(reaction: Reaction, user: User | Member):
    """Called when a message reaction is removed."""
    print(
        f"Reaction '{reaction.emoji}' removed by {user.username} "
        f"from message ID {reaction.message_id} in channel ID {reaction.channel_id}"
    )


# --- Main Execution ---
async def main():
    print("Starting Reaction Bot...")
    try:
        client.add_cog(ReactionCog(client))
        await client.run()
    except disagreement.AuthenticationError:
        print("Authentication failed. Check your bot token.")
    except Exception as e:
        print(f"An unexpected error occurred: {e}")
        traceback.print_exc()
    finally:
        if not client.is_closed():
            await client.close()
        print("Bot has been shut down.")


if __name__ == "__main__":
    asyncio.run(main())



import inspect
import json
import logging
import os
from typing import (
    TYPE_CHECKING,
    Dict,
    Optional,
    List,
    Any,
    Tuple,
    Union,
    get_origin,
    get_args,
    Literal,
)

if TYPE_CHECKING:
    from disagreement.client import Client
    from disagreement.interactions import Interaction, ResolvedData, Snowflake
    from disagreement.enums import (
        ApplicationCommandType,
        ApplicationCommandOptionType,
        InteractionType,
    )
    from .commands import (
        AppCommand,
        SlashCommand,
        UserCommand,
        MessageCommand,
        AppCommandGroup,
    )
    from .context import AppCommandContext
    from disagreement.models import (
        User,
        Member,
        Role,
        Attachment,
        Message,
    )  # For resolved data

    # Channel models would also go here

# Placeholder for models not yet fully defined or imported
if not TYPE_CHECKING:
    from disagreement.enums import (
        ApplicationCommandType,
        ApplicationCommandOptionType,
        InteractionType,
    )
    from .commands import (
        AppCommand,
        SlashCommand,
        UserCommand,
        MessageCommand,
        AppCommandGroup,
    )
    from .context import AppCommandContext

    User = Any
    Member = Any
    Role = Any
    Attachment = Any
    Channel = Any
    Message = Any


logger = logging.getLogger(__name__)

COMMANDS_CACHE_FILE = ".disagreement_commands.json"


class AppCommandHandler:
    """
    Manages application command registration, parsing, and dispatching.
    """

    def __init__(self, client: "Client"):
        self.client: "Client" = client
        # Store commands: key could be (name, type) for global, or (name, type, guild_id) for guild-specific
        # For simplicity, let's start with a flat structure and refine if needed for guild commands.
        # A more robust system might have separate dicts for global and guild commands.
        self._slash_commands: Dict[str, SlashCommand] = {}
        self._user_commands: Dict[str, UserCommand] = {}
        self._message_commands: Dict[str, MessageCommand] = {}
        self._app_command_groups: Dict[str, AppCommandGroup] = {}
        self._converter_registry: Dict[type, type] = {}

    def _load_cached_ids(self) -> Dict[str, Dict[str, str]]:
        try:
            with open(COMMANDS_CACHE_FILE, "r", encoding="utf-8") as fp:
                return json.load(fp)
        except FileNotFoundError:
            return {}
        except json.JSONDecodeError:
            logger.warning("Invalid command cache file. Ignoring.")
            return {}

    def _save_cached_ids(self, data: Dict[str, Dict[str, str]]) -> None:
        try:
            with open(COMMANDS_CACHE_FILE, "w", encoding="utf-8") as fp:
                json.dump(data, fp, indent=2)
        except Exception as e:  # pragma: no cover - logging only
            logger.error("Failed to write command cache: %s", e)

    def clear_stored_registrations(self) -> None:
        """Remove persisted command registration data."""
        if os.path.exists(COMMANDS_CACHE_FILE):
            os.remove(COMMANDS_CACHE_FILE)

    def migrate_stored_registrations(self, new_path: str) -> None:
        """Move stored registrations to ``new_path``."""
        if os.path.exists(COMMANDS_CACHE_FILE):
            os.replace(COMMANDS_CACHE_FILE, new_path)

    def add_command(self, command: Union["AppCommand", "AppCommandGroup"]) -> None:
        """Adds an application command or a command group to the handler."""
        if isinstance(command, AppCommandGroup):
            if command.name in self._app_command_groups:
                raise ValueError(
                    f"AppCommandGroup '{command.name}' is already registered."
                )
            self._app_command_groups[command.name] = command
            return

        if isinstance(command, SlashCommand):
            if command.name in self._slash_commands:
                raise ValueError(
                    f"SlashCommand '{command.name}' is already registered."
                )
            self._slash_commands[command.name] = command
            return

        if isinstance(command, UserCommand):
            if command.name in self._user_commands:
                raise ValueError(f"UserCommand '{command.name}' is already registered.")
            self._user_commands[command.name] = command
            return

        if isinstance(command, MessageCommand):
            if command.name in self._message_commands:
                raise ValueError(
                    f"MessageCommand '{command.name}' is already registered."
                )
            self._message_commands[command.name] = command
            return

        if isinstance(command, AppCommand):
            # Fallback for plain AppCommand objects
            if command.type == ApplicationCommandType.CHAT_INPUT:
                if command.name in self._slash_commands:
                    raise ValueError(
                        f"SlashCommand '{command.name}' is already registered."
                    )
                self._slash_commands[command.name] = command  # type: ignore
            elif command.type == ApplicationCommandType.USER:
                if command.name in self._user_commands:
                    raise ValueError(
                        f"UserCommand '{command.name}' is already registered."
                    )
                self._user_commands[command.name] = command  # type: ignore
            elif command.type == ApplicationCommandType.MESSAGE:
                if command.name in self._message_commands:
                    raise ValueError(
                        f"MessageCommand '{command.name}' is already registered."
                    )
                self._message_commands[command.name] = command  # type: ignore
            else:
                raise TypeError(
                    f"Unsupported command type: {command.type} for '{command.name}'"
                )
        else:
            raise TypeError("Can only add AppCommand or AppCommandGroup instances.")

    def remove_command(
        self, name: str
    ) -> Optional[Union["AppCommand", "AppCommandGroup"]]:
        """Removes an application command or group by name."""
        if name in self._slash_commands:
            return self._slash_commands.pop(name)
        if name in self._user_commands:
            return self._user_commands.pop(name)
        if name in self._message_commands:
            return self._message_commands.pop(name)
        if name in self._app_command_groups:
            return self._app_command_groups.pop(name)
        return None

    def register_converter(self, annotation: type, converter_cls: type) -> None:
        """Register a custom converter class for a type annotation."""
        self._converter_registry[annotation] = converter_cls

    def get_converter(self, annotation: type) -> Optional[type]:
        """Retrieve a registered converter class for a type annotation."""
        return self._converter_registry.get(annotation)

    def get_command(
        self,
        name: str,
        command_type: "ApplicationCommandType",
        interaction_options: Optional[List[Dict[str, Any]]] = None,
    ) -> Optional["AppCommand"]:
        """Retrieves a command of a specific type."""
        if command_type == ApplicationCommandType.CHAT_INPUT:
            if not interaction_options:
                return self._slash_commands.get(name)

            # Handle subcommands/groups
            current_options = interaction_options
            target_command_or_group: Optional[Union[AppCommand, AppCommandGroup]] = (
                self._app_command_groups.get(name)
            )

            if not target_command_or_group:
                return self._slash_commands.get(name)

            final_command: Optional[AppCommand] = None

            while current_options:
                opt_data = current_options[0]
                opt_name = opt_data.get("name")
                opt_type = (
                    ApplicationCommandOptionType(opt_data["type"])
                    if opt_data.get("type")
                    else None
                )

                if not opt_name or not isinstance(
                    target_command_or_group, AppCommandGroup
                ):
                    break

                next_target = target_command_or_group.get_command(opt_name)

                if isinstance(next_target, AppCommand) and (
                    opt_type == ApplicationCommandOptionType.SUB_COMMAND
                    or not opt_data.get("options")
                ):
                    final_command = next_target
                    break
                elif (
                    isinstance(next_target, AppCommandGroup)
                    and opt_type == ApplicationCommandOptionType.SUB_COMMAND_GROUP
                ):
                    target_command_or_group = next_target
                    current_options = opt_data.get("options", [])
                    if not current_options:
                        break
                else:
                    break

            return final_command

        if command_type == ApplicationCommandType.USER:
            return self._user_commands.get(name)

        if command_type == ApplicationCommandType.MESSAGE:
            return self._message_commands.get(name)

        return None

    async def _resolve_option_value(
        self,
        value: Any,
        expected_type: Any,
        resolved_data: Optional["ResolvedData"],
        guild_id: Optional["Snowflake"],
    ) -> Any:
        """
        Resolves an option value to the expected Python type using resolved_data.
        """
        converter_cls = self.get_converter(expected_type)
        if converter_cls:
            try:
                init_params = inspect.signature(converter_cls.__init__).parameters
                if "client" in init_params:
                    converter_instance = converter_cls(client=self.client)  # type: ignore[arg-type]
                else:
                    converter_instance = converter_cls()
                return await converter_instance.convert(None, value)  # type: ignore[arg-type]
            except Exception:
                pass

        # This is a simplified resolver. A more robust one would use converters.
        if resolved_data:
            if expected_type is User or expected_type.__name__ == "User":
                return resolved_data.users.get(value) if resolved_data.users else None

            if expected_type is Member or expected_type.__name__ == "Member":
                member_obj = (
                    resolved_data.members.get(value) if resolved_data.members else None
                )
                if member_obj:
                    if (
                        hasattr(member_obj, "username")
                        and not member_obj.username
                        and resolved_data.users
                    ):
                        user_obj = resolved_data.users.get(value)
                        if user_obj:
                            member_obj.username = user_obj.username
                            member_obj.discriminator = user_obj.discriminator
                            member_obj.avatar = user_obj.avatar
                            member_obj.bot = user_obj.bot
                            member_obj.user = user_obj  # type: ignore[attr-defined]
                    return member_obj
                return None
            if expected_type is Role or expected_type.__name__ == "Role":
                return resolved_data.roles.get(value) if resolved_data.roles else None
            if expected_type is Attachment or expected_type.__name__ == "Attachment":
                return (
                    resolved_data.attachments.get(value)
                    if resolved_data.attachments
                    else None
                )
            if expected_type is Message or expected_type.__name__ == "Message":
                return (
                    resolved_data.messages.get(value)
                    if resolved_data.messages
                    else None
                )
            if "Channel" in expected_type.__name__:
                return (
                    resolved_data.channels.get(value)
                    if resolved_data.channels
                    else None
                )

        # For basic types, Discord already sends them correctly (string, int, bool, float)
        if isinstance(value, expected_type):
            return value
        try:  # Attempt direct conversion for basic types if Discord sent string for int/float/bool
            if expected_type is int:
                return int(value)
            if expected_type is float:
                return float(value)
            if expected_type is bool:  # Discord sends true/false
                if isinstance(value, str):
                    return value.lower() == "true"
                return bool(value)
        except (ValueError, TypeError):
            pass
        return value  # Return as is if no specific resolution or conversion applied

    async def _resolve_value(
        self,
        value: Any,
        expected_type: Any,
        resolved_data: Optional["ResolvedData"],
        guild_id: Optional["Snowflake"],
    ) -> Any:
        """Public wrapper around ``_resolve_option_value`` used by tests."""

        return await self._resolve_option_value(
            value=value,
            expected_type=expected_type,
            resolved_data=resolved_data,
            guild_id=guild_id,
        )

    async def _parse_interaction_options(
        self,
        command_params: Dict[str, inspect.Parameter],  # From command.params
        interaction_options: Optional[List[Dict[str, Any]]],
        resolved_data: Optional["ResolvedData"],
        guild_id: Optional["Snowflake"],
    ) -> Tuple[List[Any], Dict[str, Any]]:
        """
        Parses options from an interaction payload and maps them to command function arguments.
        """
        args_list: List[Any] = []
        kwargs_dict: Dict[str, Any] = {}

        if not interaction_options:  # No options provided in interaction
            # Check if command has required params without defaults
            for name, param in command_params.items():
                if param.default == inspect.Parameter.empty:
                    # This should ideally be caught by Discord if option is marked required
                    raise ValueError(f"Missing required option: {name}")
            return args_list, kwargs_dict

        # Create a dictionary of provided options by name for easier lookup
        provided_options: Dict[str, Any] = {
            opt["name"]: opt["value"] for opt in interaction_options if "value" in opt
        }

        for name, param in command_params.items():
            if name in provided_options:
                raw_value = provided_options[name]
                expected_type = (
                    param.annotation
                    if param.annotation != inspect.Parameter.empty
                    else str
                )

                # Handle Optional[T]
                origin_type = get_origin(expected_type)
                if origin_type is Union:
                    union_args = get_args(expected_type)
                    # Assuming Optional[T] is Union[T, NoneType]
                    non_none_types = [t for t in union_args if t is not type(None)]
                    if len(non_none_types) == 1:
                        expected_type = non_none_types[0]
                    # Else, complex Union, might need more sophisticated handling or default to raw_value/str
                elif origin_type is Literal:
                    literal_args = get_args(expected_type)
                    if literal_args:
                        expected_type = type(literal_args[0])
                    else:
                        expected_type = str

                resolved_value = await self._resolve_option_value(
                    raw_value, expected_type, resolved_data, guild_id
                )

                if (
                    param.kind == inspect.Parameter.KEYWORD_ONLY
                    or param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
                ):
                    kwargs_dict[name] = resolved_value
                # Note: Slash commands don't map directly to *args. All options are named.
                # So, we'll primarily use kwargs_dict and then construct args_list based on param order if needed,
                # but Discord sends named options, so direct kwarg usage is more natural.
            elif param.default != inspect.Parameter.empty:
                kwargs_dict[name] = param.default
            else:
                # Required parameter not provided by Discord - this implies an issue with command definition
                # or Discord's validation, as Discord should enforce required options.
                raise ValueError(
                    f"Required option '{name}' not found in interaction payload."
                )

        # Populate args_list based on the order in command_params for positional arguments
        # This assumes that all args that are not keyword-only are passed positionally if present in kwargs_dict
        for name, param in command_params.items():
            if param.kind == inspect.Parameter.POSITIONAL_ONLY or (
                param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
                and name in kwargs_dict
            ):
                if name in kwargs_dict:  # Ensure it was resolved or had a default
                    args_list.append(kwargs_dict[name])
                # If it was POSITIONAL_ONLY and not in kwargs_dict, it's an error (already raised)
            elif param.kind == inspect.Parameter.VAR_POSITIONAL:  # *args
                # Slash commands don't map to *args well. This would be empty.
                pass

        # Filter kwargs_dict to only include actual KEYWORD_ONLY or POSITIONAL_OR_KEYWORD params
        # that were not used for args_list (if strict positional/keyword separation is desired).
        # For slash commands, it's simpler to pass all resolved named options as kwargs.
        final_kwargs = {
            k: v
            for k, v in kwargs_dict.items()
            if k in command_params
            and command_params[k].kind != inspect.Parameter.POSITIONAL_ONLY
        }

        # For simplicity with slash commands, let's assume all resolved options are passed via kwargs
        # and the command signature is primarily (self, ctx, **options) or (ctx, **options)
        # or (self, ctx, option1, option2) where names match.
        # The AppCommand.invoke will handle passing them.
        # The current args_list and final_kwargs might be redundant if invoke just uses **final_kwargs.
        # Let's return kwargs_dict directly for now, and AppCommand.invoke can map them.

        return [], kwargs_dict  # Return empty args, all in kwargs for now.

    async def dispatch_app_command_error(
        self, context: "AppCommandContext", error: Exception
    ) -> None:
        """Dispatches an app command error to the client if implemented."""
        if hasattr(self.client, "on_app_command_error"):
            await self.client.on_app_command_error(context, error)

    async def process_interaction(self, interaction: "Interaction") -> None:
        """Processes an incoming interaction."""
        if interaction.type == InteractionType.MODAL_SUBMIT:
            callback = getattr(self.client, "on_modal_submit", None)
            if callback is not None:
                from typing import Awaitable, Callable, cast

                await cast(Callable[["Interaction"], Awaitable[None]], callback)(
                    interaction
                )
            return

        if interaction.type == InteractionType.APPLICATION_COMMAND_AUTOCOMPLETE:
            callback = getattr(self.client, "on_autocomplete", None)
            if callback is not None:
                from typing import Awaitable, Callable, cast

                await cast(Callable[["Interaction"], Awaitable[None]], callback)(
                    interaction
                )
            return

        if interaction.type != InteractionType.APPLICATION_COMMAND:
            return

        if not interaction.data or not interaction.data.name:
            from .context import AppCommandContext

            ctx = AppCommandContext(
                bot=self.client, interaction=interaction, command=None
            )
            await ctx.send("Command not found.", ephemeral=True)
            return

        command_name = interaction.data.name
        command_type = interaction.data.type or ApplicationCommandType.CHAT_INPUT
        command = self.get_command(
            command_name,
            command_type,
            interaction.data.options if interaction.data else None,
        )

        if not command:
            from .context import AppCommandContext

            ctx = AppCommandContext(
                bot=self.client, interaction=interaction, command=None
            )
            await ctx.send(f"Command '{command_name}' not found.", ephemeral=True)
            return

        # Create context
        from .context import AppCommandContext  # Ensure AppCommandContext is available

        ctx = AppCommandContext(
            bot=self.client, interaction=interaction, command=command
        )

        try:
            # Prepare arguments for the command callback
            # Skip 'self' and 'ctx' from command.params for parsing interaction options
            params_to_parse = {
                name: param
                for name, param in command.params.items()
                if name not in ("self", "ctx")
            }

            if command.type in (
                ApplicationCommandType.USER,
                ApplicationCommandType.MESSAGE,
            ):
                # Context menu commands provide a target_id. Resolve and pass it
                args = []
                kwargs = {}
                if params_to_parse and interaction.data and interaction.data.target_id:
                    first_param = next(iter(params_to_parse.values()))
                    expected = (
                        first_param.annotation
                        if first_param.annotation != inspect.Parameter.empty
                        else str
                    )
                    resolved = await self._resolve_option_value(
                        interaction.data.target_id,
                        expected,
                        interaction.data.resolved,
                        interaction.guild_id,
                    )
                    if first_param.kind in (
                        inspect.Parameter.POSITIONAL_ONLY,
                        inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    ):
                        args.append(resolved)
                    else:
                        kwargs[first_param.name] = resolved

                await command.invoke(ctx, *args, **kwargs)
            else:
                parsed_args, parsed_kwargs = await self._parse_interaction_options(
                    command_params=params_to_parse,
                    interaction_options=interaction.data.options,
                    resolved_data=interaction.data.resolved,
                    guild_id=interaction.guild_id,
                )

                await command.invoke(ctx, *parsed_args, **parsed_kwargs)

        except Exception as e:
            logger.error("Error invoking app command '%s': %s", command.name, e)
            await self.dispatch_app_command_error(ctx, e)
            # else:
            #     # Default error reply if no handler on client
            #     try:
            #         await ctx.send(f"An error occurred: {e}", ephemeral=True)
            #     except Exception as send_e:
            #         print(f"Failed to send error message for app command: {send_e}")

    async def sync_commands(
        self, application_id: "Snowflake", guild_id: Optional["Snowflake"] = None
    ) -> None:
        """
        Synchronizes (registers/updates) all application commands with Discord.
        If guild_id is provided, syncs commands for that guild. Otherwise, syncs global commands.
        """
        cache = self._load_cached_ids()
        scope_key = str(guild_id) if guild_id else "global"
        stored = cache.get(scope_key, {})

        current_payloads: Dict[str, Dict[str, Any]] = {}

        # Collect commands based on scope (global or specific guild)
        source_commands = (
            list(self._slash_commands.values())
            + list(self._user_commands.values())
            + list(self._message_commands.values())
            + list(self._app_command_groups.values())
        )

        for cmd_or_group in source_commands:
            is_guild_specific_command = (
                cmd_or_group.guild_ids is not None and len(cmd_or_group.guild_ids) > 0
            )

            if guild_id:
                if not is_guild_specific_command or (
                    cmd_or_group.guild_ids is not None
                    and guild_id not in cmd_or_group.guild_ids
                ):
                    continue
            else:
                if is_guild_specific_command:
                    continue

            try:
                current_payloads[cmd_or_group.name] = cmd_or_group.to_dict()
            except AttributeError:
                logger.warning(
                    "Command or group '%s' does not have a to_dict() method. Skipping.",
                    cmd_or_group.name,
                )
            except Exception as e:
                logger.error(
                    "Error converting command/group '%s' to dict: %s. Skipping.",
                    cmd_or_group.name,
                    e,
                )

        if not current_payloads:
            logger.info(
                "No commands to sync for %s scope.",
                f"guild {guild_id}" if guild_id else "global",
            )
            return

        names_current = set(current_payloads)
        names_stored = set(stored)

        to_delete = names_stored - names_current
        to_create = names_current - names_stored
        to_update = names_current & names_stored

        if not to_delete and not to_create and not to_update:
            logger.info(
                "Application commands already up to date for %s scope.", scope_key
            )
            return

        try:
            for name in to_delete:
                cmd_id = stored[name]
                if guild_id:
                    await self.client._http.delete_guild_application_command(
                        application_id, guild_id, cmd_id
                    )
                else:
                    await self.client._http.delete_global_application_command(
                        application_id, cmd_id
                    )

            new_ids: Dict[str, str] = {}
            for name in to_create:
                payload = current_payloads[name]
                if guild_id:
                    result = await self.client._http.create_guild_application_command(
                        application_id, guild_id, payload
                    )
                else:
                    result = await self.client._http.create_global_application_command(
                        application_id, payload
                    )
                if result.id:
                    new_ids[name] = str(result.id)

            for name in to_update:
                payload = current_payloads[name]
                cmd_id = stored[name]
                if guild_id:
                    await self.client._http.edit_guild_application_command(
                        application_id, guild_id, cmd_id, payload
                    )
                else:
                    await self.client._http.edit_global_application_command(
                        application_id, cmd_id, payload
                    )
                new_ids[name] = cmd_id

            final_ids: Dict[str, str] = {}
            for name in names_current:
                if name in new_ids:
                    final_ids[name] = new_ids[name]
                else:
                    final_ids[name] = stored[name]

            cache[scope_key] = final_ids
            self._save_cached_ids(cache)
            logger.info("Command sync successful.")
        except Exception as e:
            logger.error("Error syncing application commands: %s", e)



import pytest
from unittest.mock import AsyncMock

from disagreement.client import Client
from disagreement.http import HTTPClient
from disagreement.models import Attachment, File


@pytest.mark.asyncio
async def test_http_send_message_includes_attachments():
    http = HTTPClient(token="t")
    http.request = AsyncMock(
        return_value={
            "id": "1",
            "channel_id": "c",
            "author": {"id": "2", "username": "u", "discriminator": "0001"},
            "content": "hi",
            "timestamp": "t",
        }
    )
    await http.send_message("c", "hi", attachments=[{"id": 1}])
    http.request.assert_called_once_with(
        "POST",
        "/channels/c/messages",
        payload={"content": "hi", "attachments": [{"id": 1}]},
    )


@pytest.mark.asyncio
async def test_http_send_message_with_files_uses_formdata():
    http = HTTPClient(token="t")
    http.request = AsyncMock(
        return_value={
            "id": "1",
            "channel_id": "c",
            "author": {"id": "2", "username": "u", "discriminator": "0001"},
            "content": "hi",
            "timestamp": "t",
        }
    )
    await http.send_message("c", "hi", files=[File("f.txt", b"data")])
    args, kwargs = http.request.call_args
    assert kwargs["is_json"] is False


@pytest.mark.asyncio
async def test_client_send_message_passes_attachments():
    client = Client(token="t")
    client._http.send_message = AsyncMock(
        return_value={
            "id": "1",
            "channel_id": "c",
            "author": {"id": "2", "username": "u", "discriminator": "0001"},
            "content": "hi",
            "timestamp": "t",
            "attachments": [{"id": "1", "filename": "f.txt"}],
        }
    )
    msg = await client.send_message("c", "hi", attachments=[{"id": "1"}])
    client._http.send_message.assert_awaited_once()
    kwargs = client._http.send_message.call_args.kwargs
    assert kwargs["attachments"] == [{"id": "1"}]
    assert isinstance(msg.attachments[0], Attachment)


@pytest.mark.asyncio
async def test_client_send_message_passes_files():
    client = Client(token="t")
    client._http.send_message = AsyncMock(
        return_value={
            "id": "1",
            "channel_id": "c",
            "author": {"id": "2", "username": "u", "discriminator": "0001"},
            "content": "hi",
            "timestamp": "t",
        }
    )
    await client.send_message("c", "hi", files=[File("f.txt", b"data")])
    client._http.send_message.assert_awaited_once()
    kwargs = client._http.send_message.call_args.kwargs
    assert kwargs["files"][0].filename == "f.txt"



import asyncio
import datetime
from typing import Any, Awaitable, Callable, Optional

__all__ = ["loop", "Task"]


class Task:
    """Simple repeating task."""

    def __init__(
        self,
        coro: Callable[..., Awaitable[Any]],
        *,
        seconds: float = 0.0,
        minutes: float = 0.0,
        hours: float = 0.0,
        delta: Optional[datetime.timedelta] = None,
        time_of_day: Optional[datetime.time] = None,
        on_error: Optional[Callable[[Exception], Awaitable[None]]] = None,
        before_loop: Optional[Callable[[], Awaitable[None] | None]] = None,
        after_loop: Optional[Callable[[], Awaitable[None] | None]] = None,
    ) -> None:
        self._coro = coro
        self._task: Optional[asyncio.Task[None]] = None
        if time_of_day is not None and (
            seconds or minutes or hours or delta is not None
        ):
            raise ValueError("time_of_day cannot be used with an interval")

        if delta is not None:
            if not isinstance(delta, datetime.timedelta):
                raise TypeError("delta must be a datetime.timedelta")
            interval_seconds = delta.total_seconds()
        else:
            interval_seconds = seconds + minutes * 60.0 + hours * 3600.0

        self._seconds = float(interval_seconds)
        self._time_of_day = time_of_day
        self._on_error = on_error
        self._before_loop = before_loop
        self._after_loop = after_loop

    def _seconds_until_time(self) -> float:
        assert self._time_of_day is not None
        now = datetime.datetime.now()
        target = datetime.datetime.combine(now.date(), self._time_of_day)
        if target <= now:
            target += datetime.timedelta(days=1)
        return (target - now).total_seconds()

    async def _run(self, *args: Any, **kwargs: Any) -> None:
        try:
            if self._before_loop is not None:
                await _maybe_call_no_args(self._before_loop)

            first = True
            while True:
                if self._time_of_day is not None:
                    await asyncio.sleep(self._seconds_until_time())
                elif not first:
                    await asyncio.sleep(self._seconds)

                try:
                    await self._coro(*args, **kwargs)
                except Exception as exc:  # noqa: BLE001
                    if self._on_error is not None:
                        await _maybe_call(self._on_error, exc)
                    else:
                        raise

                first = False
        except asyncio.CancelledError:
            pass
        finally:
            if self._after_loop is not None:
                await _maybe_call_no_args(self._after_loop)

    def start(self, *args: Any, **kwargs: Any) -> asyncio.Task[None]:
        if self._task is None or self._task.done():
            self._task = asyncio.create_task(self._run(*args, **kwargs))
        return self._task

    def stop(self) -> None:
        if self._task is not None:
            self._task.cancel()
            self._task = None

    @property
    def running(self) -> bool:
        return self._task is not None and not self._task.done()


async def _maybe_call(
    func: Callable[[Exception], Awaitable[None] | None], exc: Exception
) -> None:
    result = func(exc)
    if asyncio.iscoroutine(result):
        await result


async def _maybe_call_no_args(func: Callable[[], Awaitable[None] | None]) -> None:
    result = func()
    if asyncio.iscoroutine(result):
        await result


class _Loop:
    def __init__(
        self,
        func: Callable[..., Awaitable[Any]],
        *,
        seconds: float = 0.0,
        minutes: float = 0.0,
        hours: float = 0.0,
        delta: Optional[datetime.timedelta] = None,
        time_of_day: Optional[datetime.time] = None,
        on_error: Optional[Callable[[Exception], Awaitable[None]]] = None,
    ) -> None:
        self.func = func
        self.seconds = seconds
        self.minutes = minutes
        self.hours = hours
        self.delta = delta
        self.time_of_day = time_of_day
        self.on_error = on_error
        self._task: Optional[Task] = None
        self._owner: Any = None
        self._before_loop: Optional[Callable[..., Awaitable[Any]]] = None
        self._after_loop: Optional[Callable[..., Awaitable[Any]]] = None

    def __get__(self, obj: Any, objtype: Any) -> "_BoundLoop":
        return _BoundLoop(self, obj)

    def _coro(self, *args: Any, **kwargs: Any) -> Awaitable[Any]:
        if self._owner is None:
            return self.func(*args, **kwargs)
        return self.func(self._owner, *args, **kwargs)

    def before_loop(
        self, func: Callable[..., Awaitable[Any]]
    ) -> Callable[..., Awaitable[Any]]:
        self._before_loop = func
        return func

    def after_loop(
        self, func: Callable[..., Awaitable[Any]]
    ) -> Callable[..., Awaitable[Any]]:
        self._after_loop = func
        return func

    def start(self, *args: Any, **kwargs: Any) -> asyncio.Task[None]:
        def call_before() -> Awaitable[None] | None:
            if self._before_loop is None:
                return None
            if self._owner is not None:
                return self._before_loop(self._owner)
            return self._before_loop()

        def call_after() -> Awaitable[None] | None:
            if self._after_loop is None:
                return None
            if self._owner is not None:
                return self._after_loop(self._owner)
            return self._after_loop()

        self._task = Task(
            self._coro,
            seconds=self.seconds,
            minutes=self.minutes,
            hours=self.hours,
            delta=self.delta,
            time_of_day=self.time_of_day,
            on_error=self.on_error,
            before_loop=call_before,
            after_loop=call_after,
        )
        return self._task.start(*args, **kwargs)

    def stop(self) -> None:
        if self._task is not None:
            self._task.stop()

    @property
    def running(self) -> bool:
        return self._task.running if self._task else False


class _BoundLoop:
    def __init__(self, parent: _Loop, owner: Any) -> None:
        self._parent = parent
        self._owner = owner

    def start(self, *args: Any, **kwargs: Any) -> asyncio.Task[None]:
        self._parent._owner = self._owner
        return self._parent.start(*args, **kwargs)

    def stop(self) -> None:
        self._parent.stop()

    @property
    def running(self) -> bool:
        return self._parent.running


def loop(
    *,
    seconds: float = 0.0,
    minutes: float = 0.0,
    hours: float = 0.0,
    delta: Optional[datetime.timedelta] = None,
    time_of_day: Optional[datetime.time] = None,
    on_error: Optional[Callable[[Exception], Awaitable[None]]] = None,
) -> Callable[[Callable[..., Awaitable[Any]]], _Loop]:
    """Decorator to create a looping task."""

    def decorator(func: Callable[..., Awaitable[Any]]) -> _Loop:
        return _Loop(
            func,
            seconds=seconds,
            minutes=minutes,
            hours=hours,
            delta=delta,
            time_of_day=time_of_day,
            on_error=on_error,
        )

    return decorator



import asyncio
import pytest

from disagreement.ext.commands.core import CommandHandler
from disagreement.ext.commands.decorators import command, max_concurrency
from disagreement.ext.commands.errors import MaxConcurrencyReached
from disagreement.models import Message


class DummyBot:
    def __init__(self):
        self.errors = []

    async def on_command_error(self, ctx, error):
        self.errors.append(error)


@pytest.mark.asyncio
async def test_max_concurrency_per_user():
    bot = DummyBot()
    handler = CommandHandler(client=bot, prefix="!")
    started = asyncio.Event()
    release = asyncio.Event()

    @command()
    @max_concurrency(1, per="user")
    async def foo(ctx):
        started.set()
        await release.wait()

    handler.add_command(foo.__command_object__)

    data = {
        "id": "1",
        "channel_id": "c",
        "guild_id": "g",
        "author": {"id": "a", "username": "u", "discriminator": "0001"},
        "content": "!foo",
        "timestamp": "t",
    }
    msg1 = Message(data, client_instance=bot)
    msg2 = Message({**data, "id": "2"}, client_instance=bot)

    task = asyncio.create_task(handler.process_commands(msg1))
    await started.wait()

    await handler.process_commands(msg2)
    assert any(isinstance(e, MaxConcurrencyReached) for e in bot.errors)

    release.set()
    await task

    await handler.process_commands(msg2)


@pytest.mark.asyncio
async def test_max_concurrency_per_guild():
    bot = DummyBot()
    handler = CommandHandler(client=bot, prefix="!")
    started = asyncio.Event()
    release = asyncio.Event()

    @command()
    @max_concurrency(1, per="guild")
    async def foo(ctx):
        started.set()
        await release.wait()

    handler.add_command(foo.__command_object__)

    base = {
        "channel_id": "c",
        "guild_id": "g",
        "content": "!foo",
        "timestamp": "t",
    }
    msg1 = Message(
        {
            **base,
            "id": "1",
            "author": {"id": "a", "username": "u", "discriminator": "0001"},
        },
        client_instance=bot,
    )
    msg2 = Message(
        {
            **base,
            "id": "2",
            "author": {"id": "b", "username": "v", "discriminator": "0001"},
        },
        client_instance=bot,
    )

    task = asyncio.create_task(handler.process_commands(msg1))
    await started.wait()

    await handler.process_commands(msg2)
    assert any(isinstance(e, MaxConcurrencyReached) for e in bot.errors)

    release.set()
    await task

    await handler.process_commands(msg2)



from typing import Any, Callable, List, Optional

from .commands import SlashCommand
from disagreement.ext.commands.core import PrefixCommand


class HybridCommand(SlashCommand, PrefixCommand):  # Inherit from both
    """
    Represents a command that can be invoked as both a slash command
    and a traditional prefix-based command.
    """

    def __init__(self, callback: Callable[..., Any], **kwargs: Any):
        # Initialize SlashCommand part (which calls AppCommand.__init__)
        # We need to ensure 'type' is correctly passed for AppCommand
        # kwargs for SlashCommand: name, description, guild_ids, default_member_permissions, nsfw, parent, cog, etc.
        # kwargs for PrefixCommand: name, aliases, brief, description, cog

        # Pop prefix-specific args before passing to SlashCommand constructor
        prefix_aliases = kwargs.pop("aliases", [])
        prefix_brief = kwargs.pop("brief", None)
        # Description is used by both, AppCommand's constructor will handle it.
        # Name is used by both. Cog is used by both.

        # Call SlashCommand's __init__
        # This will set up name, description, callback, type=CHAT_INPUT, options, etc.
        super().__init__(callback, **kwargs)  # This is SlashCommand.__init__

        # Now, explicitly initialize the PrefixCommand parts that SlashCommand didn't cover
        # or that need specific values for the prefix version.
        # PrefixCommand.__init__(self, callback, name=self.name, aliases=prefix_aliases, brief=prefix_brief, description=self.description, cog=self.cog)
        # However, PrefixCommand.__init__ also sets self.params, which AppCommand already did.
        # We need to be careful not to re-initialize things unnecessarily or incorrectly.
        # Let's manually set the distinct attributes for the PrefixCommand aspect.

        # Attributes from PrefixCommand:
        # self.callback is already set by AppCommand
        # self.name is already set by AppCommand
        self.aliases: List[str] = (
            prefix_aliases  # This was specific to HybridCommand before, now aligns with PrefixCommand
        )
        self.brief: Optional[str] = prefix_brief
        # self.description is already set by AppCommand (SlashCommand ensures it exists)
        # self.cog is already set by AppCommand
        # self.params is already set by AppCommand

        # Ensure the MRO is handled correctly. Python's MRO (C3 linearization)
        # should call SlashCommand's __init__ then AppCommand's __init__.
        # PrefixCommand.__init__ won't be called automatically unless we explicitly call it.
        # By setting attributes directly, we avoid potential issues with multiple __init__ calls
        # if their logic overlaps too much (e.g., both trying to set self.params).

    # We might need to override invoke if the context or argument passing differs significantly
    # between app command invocation and prefix command invocation.
    # For now, SlashCommand.invoke and PrefixCommand.invoke are separate.
    # The correct one will be called depending on how the command is dispatched.
    # The AppCommandHandler will use AppCommand.invoke (via SlashCommand).
    # The prefix CommandHandler will use PrefixCommand.invoke.
    # This seems acceptable.



import inspect
from typing import Any, Callable, Dict, List, Optional, Union, TYPE_CHECKING

from disagreement.enums import (
    ApplicationCommandType,
    ApplicationCommandOptionType,
    IntegrationType,
    InteractionContextType,
)
from disagreement.interactions import ApplicationCommandOption, Snowflake

if TYPE_CHECKING:
    from disagreement.ext.commands.core import Command as PrefixCommand
    from disagreement.ext.commands.cog import Cog

if not TYPE_CHECKING:
    try:
        from disagreement.ext.commands.core import Command as PrefixCommand
    except ImportError:
        PrefixCommand = Any


class AppCommand:
    """
    Base class for an application command.

    Attributes:
        name (str): The name of the command.
        description (Optional[str]): The description of the command.
            Required for CHAT_INPUT, empty for USER and MESSAGE commands.
        callback (Callable[..., Any]): The coroutine function that will be called when the command is executed.
        type (ApplicationCommandType): The type of the application command.
        options (Optional[List[ApplicationCommandOption]]): Parameters for the command. Populated by decorators.
        guild_ids (Optional[List[Snowflake]]): List of guild IDs where this command is active. None for global.
        default_member_permissions (Optional[str]): Bitwise permissions required by default for users to run the command.
        nsfw (bool): Whether the command is age-restricted.
        parent (Optional['AppCommandGroup']): The parent group if this is a subcommand.
        cog (Optional[Cog]): The cog this command belongs to, if any.
        _full_description (Optional[str]): Stores the original full description, e.g. from docstring,
                                          even if the payload description is different (like for User/Message commands).
        name_localizations (Optional[Dict[str, str]]): Localizations for the command's name.
        description_localizations (Optional[Dict[str, str]]): Localizations for the command's description.
        integration_types (Optional[List[IntegrationType]]): Installation contexts.
        contexts (Optional[List[InteractionContextType]]): Interaction contexts.
    """

    def __init__(
        self,
        callback: Callable[..., Any],
        *,
        name: str,
        description: Optional[str] = None,
        locale: Optional[str] = None,
        type: "ApplicationCommandType",
        guild_ids: Optional[List["Snowflake"]] = None,
        default_member_permissions: Optional[str] = None,
        nsfw: bool = False,
        parent: Optional["AppCommandGroup"] = None,
        cog: Optional[
            Any
        ] = None,  # Changed 'Cog' to Any to avoid runtime import issues if Cog is complex
        name_localizations: Optional[Dict[str, str]] = None,
        description_localizations: Optional[Dict[str, str]] = None,
        integration_types: Optional[List["IntegrationType"]] = None,
        contexts: Optional[List["InteractionContextType"]] = None,
    ):
        if not asyncio.iscoroutinefunction(callback):
            raise TypeError(
                "Application command callback must be a coroutine function."
            )

        if locale:
            from disagreement import i18n

            translate = i18n.translate

            self.name = translate(name, locale)
            self.description = (
                translate(description, locale) if description is not None else None
            )
        else:
            self.name = name
            self.description = description
        self.locale: Optional[str] = locale
        self.callback: Callable[..., Any] = callback
        self.type: "ApplicationCommandType" = type
        self.options: List["ApplicationCommandOption"] = []  # Populated by decorator
        self.guild_ids: Optional[List["Snowflake"]] = guild_ids
        self.default_member_permissions: Optional[str] = default_member_permissions
        self.nsfw: bool = nsfw
        self.parent: Optional["AppCommandGroup"] = parent
        self.cog: Optional[Any] = cog  # Changed 'Cog' to Any
        self.name_localizations: Optional[Dict[str, str]] = name_localizations
        self.description_localizations: Optional[Dict[str, str]] = (
            description_localizations
        )
        self.integration_types: Optional[List["IntegrationType"]] = integration_types
        self.contexts: Optional[List["InteractionContextType"]] = contexts
        self._full_description: Optional[str] = (
            None  # Initialized by decorator if needed
        )

        # Signature for argument parsing by decorators/handlers
        self.params = inspect.signature(callback).parameters

    async def invoke(
        self, context: "AppCommandContext", *args: Any, **kwargs: Any
    ) -> None:
        """Invokes the command's callback with the given context and arguments."""
        # Similar to Command.invoke, handle cog if present
        actual_args = []
        if self.cog:
            actual_args.append(self.cog)
        actual_args.append(context)
        actual_args.extend(args)

        await self.callback(*actual_args, **kwargs)

    def to_dict(self) -> Dict[str, Any]:
        """Converts the command to a dictionary payload for Discord API."""
        payload: Dict[str, Any] = {
            "name": self.name,
            "type": self.type.value,
            # CHAT_INPUT commands require a description.
            # USER and MESSAGE commands must have an empty description in the payload if not omitted.
            # The constructor for UserCommand/MessageCommand already sets self.description to ""
            "description": (
                self.description
                if self.type == ApplicationCommandType.CHAT_INPUT
                else ""
            ),
        }

        # For CHAT_INPUT commands, options are its parameters.
        # For USER/MESSAGE commands, options should be empty or not present.
        if self.type == ApplicationCommandType.CHAT_INPUT and self.options:
            payload["options"] = [opt.to_dict() for opt in self.options]

        if self.default_member_permissions is not None:  # Can be "0" for no permissions
            payload["default_member_permissions"] = str(self.default_member_permissions)

        # nsfw defaults to False, only include if True
        if self.nsfw:
            payload["nsfw"] = True

        if self.name_localizations:
            payload["name_localizations"] = self.name_localizations

        # Description localizations only apply if there's a description (CHAT_INPUT commands)
        if (
            self.type == ApplicationCommandType.CHAT_INPUT
            and self.description
            and self.description_localizations
        ):
            payload["description_localizations"] = self.description_localizations

        if self.integration_types:
            payload["integration_types"] = [it.value for it in self.integration_types]

        if self.contexts:
            payload["contexts"] = [ict.value for ict in self.contexts]

        # According to Discord API, guild_id is not part of this payload,
        # it's used in the URL path for guild-specific command registration.
        # However, the global command registration takes an 'application_id' in the payload,
        # but that's handled by the HTTPClient.

        return payload

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name='{self.name}' type={self.type!r}>"


class SlashCommand(AppCommand):
    """Represents a CHAT_INPUT (slash) command."""

    def __init__(self, callback: Callable[..., Any], **kwargs: Any):
        if not kwargs.get("description"):
            raise ValueError("SlashCommand requires a description.")
        super().__init__(callback, type=ApplicationCommandType.CHAT_INPUT, **kwargs)


class UserCommand(AppCommand):
    """Represents a USER context menu command."""

    def __init__(self, callback: Callable[..., Any], **kwargs: Any):
        # Description is not allowed by Discord API for User Commands, but can be set to empty string.
        kwargs["description"] = kwargs.get(
            "description", ""
        )  # Ensure it's empty or not present in payload
        super().__init__(callback, type=ApplicationCommandType.USER, **kwargs)


class MessageCommand(AppCommand):
    """Represents a MESSAGE context menu command."""

    def __init__(self, callback: Callable[..., Any], **kwargs: Any):
        # Description is not allowed by Discord API for Message Commands.
        kwargs["description"] = kwargs.get("description", "")
        super().__init__(callback, type=ApplicationCommandType.MESSAGE, **kwargs)


class AppCommandGroup:
    """
    Represents a group of application commands (subcommands or subcommand groups).
    This itself is not directly callable but acts as a namespace.
    """

    def __init__(
        self,
        name: str,
        description: Optional[
            str
        ] = None,  # Required for top-level groups that form part of a slash command
        guild_ids: Optional[List["Snowflake"]] = None,
        parent: Optional["AppCommandGroup"] = None,
        default_member_permissions: Optional[str] = None,
        nsfw: bool = False,
        name_localizations: Optional[Dict[str, str]] = None,
        description_localizations: Optional[Dict[str, str]] = None,
        integration_types: Optional[List["IntegrationType"]] = None,
        contexts: Optional[List["InteractionContextType"]] = None,
    ):
        self.name: str = name
        self.description: Optional[str] = description
        self.guild_ids: Optional[List["Snowflake"]] = guild_ids
        self.parent: Optional["AppCommandGroup"] = parent
        self.commands: Dict[str, Union[AppCommand, "AppCommandGroup"]] = {}
        self.default_member_permissions: Optional[str] = default_member_permissions
        self.nsfw: bool = nsfw
        self.name_localizations: Optional[Dict[str, str]] = name_localizations
        self.description_localizations: Optional[Dict[str, str]] = (
            description_localizations
        )
        self.integration_types: Optional[List["IntegrationType"]] = integration_types
        self.contexts: Optional[List["InteractionContextType"]] = contexts
        # A group itself doesn't have a cog directly, its commands do.

    def add_command(self, command: Union[AppCommand, "AppCommandGroup"]) -> None:
        if command.name in self.commands:
            raise ValueError(
                f"Command or group '{command.name}' already exists in group '{self.name}'."
            )
        command.parent = self
        self.commands[command.name] = command

    def get_command(self, name: str) -> Optional[Union[AppCommand, "AppCommandGroup"]]:
        return self.commands.get(name)

    def command(self, *d_args: Any, **d_kwargs: Any):
        d_kwargs.setdefault("parent", self)
        from .decorators import slash_command

        return slash_command(*d_args, **d_kwargs)

    def group(
        self,
        name: str,
        description: Optional[str] = None,
        **kwargs: Any,
    ):
        sub_group = AppCommandGroup(
            name=name,
            description=description,
            parent=self,
            guild_ids=kwargs.get("guild_ids"),
            default_member_permissions=kwargs.get("default_member_permissions"),
            nsfw=kwargs.get("nsfw", False),
            name_localizations=kwargs.get("name_localizations"),
            description_localizations=kwargs.get("description_localizations"),
            integration_types=kwargs.get("integration_types"),
            contexts=kwargs.get("contexts"),
        )
        self.add_command(sub_group)

        def decorator(func: Optional[Callable[..., Any]] = None):
            if func is not None:
                setattr(func, "__app_command_object__", sub_group)
                return sub_group
            return sub_group

        return decorator

    def __repr__(self) -> str:
        return f"<AppCommandGroup name='{self.name}' commands={len(self.commands)}>"

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the command group to a dictionary payload for Discord API.
        This represents a top-level command that has subcommands/subcommand groups.
        """
        payload: Dict[str, Any] = {
            "name": self.name,
            "type": ApplicationCommandType.CHAT_INPUT.value,  # Groups are implicitly CHAT_INPUT
            "description": self.description
            or "No description provided",  # Top-level groups require a description
            "options": [],
        }

        if self.default_member_permissions is not None:
            payload["default_member_permissions"] = str(self.default_member_permissions)
        if self.nsfw:
            payload["nsfw"] = True
        if self.name_localizations:
            payload["name_localizations"] = self.name_localizations
        if (
            self.description and self.description_localizations
        ):  # Only if description is not empty
            payload["description_localizations"] = self.description_localizations
        if self.integration_types:
            payload["integration_types"] = [it.value for it in self.integration_types]
        if self.contexts:
            payload["contexts"] = [ict.value for ict in self.contexts]

        # guild_ids are handled at the registration level, not in this specific payload part.

        options_payload: List[Dict[str, Any]] = []
        for cmd_name, command_or_group in self.commands.items():
            if isinstance(command_or_group, AppCommand):  # This is a Subcommand
                # Subcommands use their own options (parameters)
                sub_options = (
                    [opt.to_dict() for opt in command_or_group.options]
                    if command_or_group.options
                    else []
                )
                option_dict = {
                    "type": ApplicationCommandOptionType.SUB_COMMAND.value,
                    "name": command_or_group.name,
                    "description": command_or_group.description
                    or "No description provided",
                    "options": sub_options,
                }
                # Add localization for subcommand name and description if available
                if command_or_group.name_localizations:
                    option_dict["name_localizations"] = (
                        command_or_group.name_localizations
                    )
                if (
                    command_or_group.description
                    and command_or_group.description_localizations
                ):
                    option_dict["description_localizations"] = (
                        command_or_group.description_localizations
                    )
                options_payload.append(option_dict)

            elif isinstance(
                command_or_group, AppCommandGroup
            ):  # This is a Subcommand Group
                # Subcommand groups have their subcommands/groups as options
                sub_group_options: List[Dict[str, Any]] = []
                for sub_cmd_name, sub_command in command_or_group.commands.items():
                    # Nested groups can only contain subcommands, not further nested groups as per Discord rules.
                    # So, sub_command here must be an AppCommand.
                    if isinstance(
                        sub_command, AppCommand
                    ):  # Should always be AppCommand if structure is valid
                        sub_cmd_options = (
                            [opt.to_dict() for opt in sub_command.options]
                            if sub_command.options
                            else []
                        )
                        sub_group_option_entry = {
                            "type": ApplicationCommandOptionType.SUB_COMMAND.value,
                            "name": sub_command.name,
                            "description": sub_command.description
                            or "No description provided",
                            "options": sub_cmd_options,
                        }
                        # Add localization for subcommand name and description if available
                        if sub_command.name_localizations:
                            sub_group_option_entry["name_localizations"] = (
                                sub_command.name_localizations
                            )
                        if (
                            sub_command.description
                            and sub_command.description_localizations
                        ):
                            sub_group_option_entry["description_localizations"] = (
                                sub_command.description_localizations
                            )
                        sub_group_options.append(sub_group_option_entry)
                    # else:
                    #     # This case implies a group nested inside a group, which then contains another group.
                    #     # Discord's structure is:
                    #     # command -> option (SUB_COMMAND_GROUP) -> option (SUB_COMMAND) -> option (param)
                    #     # This should be caught by validation logic in decorators or add_command.
                    #     # For now, we assume valid structure where AppCommandGroup's commands are AppCommands.
                    #     pass

                option_dict = {
                    "type": ApplicationCommandOptionType.SUB_COMMAND_GROUP.value,
                    "name": command_or_group.name,
                    "description": command_or_group.description
                    or "No description provided",
                    "options": sub_group_options,  # These are the SUB_COMMANDs
                }
                # Add localization for subcommand group name and description if available
                if command_or_group.name_localizations:
                    option_dict["name_localizations"] = (
                        command_or_group.name_localizations
                    )
                if (
                    command_or_group.description
                    and command_or_group.description_localizations
                ):
                    option_dict["description_localizations"] = (
                        command_or_group.description_localizations
                    )
                options_payload.append(option_dict)

        payload["options"] = options_payload
        return payload


# Need to import asyncio for iscoroutinefunction check
import asyncio

if TYPE_CHECKING:
    from .context import AppCommandContext  # For type hint in AppCommand.invoke

    # Ensure ApplicationCommandOptionType is available for the to_dict method
    from disagreement.enums import ApplicationCommandOptionType



import asyncio
from unittest.mock import AsyncMock

import pytest

from disagreement.ext.app_commands.context import AppCommandContext
from disagreement.interactions import Interaction
from disagreement.enums import InteractionType, MessageFlags
from disagreement.models import (
    Embed,
    ActionRow,
    Button,
    Container,
    TextDisplay,
)
from disagreement.enums import ButtonStyle, ComponentType


class DummyHTTP:
    def __init__(self):
        self.create_interaction_response = AsyncMock()
        self.create_followup_message = AsyncMock(return_value={"id": "123"})
        self.edit_original_interaction_response = AsyncMock(return_value={"id": "123"})
        self.edit_followup_message = AsyncMock(return_value={"id": "123"})


class DummyBot:
    def __init__(self):
        self._http = DummyHTTP()
        self.application_id = "app123"
        self._guilds = {}
        self._channels = {}

    def get_guild(self, gid):
        return self._guilds.get(gid)

    async def fetch_channel(self, cid):
        return self._channels.get(cid)


from disagreement.ext.commands.core import CommandContext, Command
from disagreement.enums import MessageFlags, ButtonStyle, ComponentType
from disagreement.models import ActionRow, Button, Container, TextDisplay


@pytest.mark.asyncio
async def test_sends_extra_payload(dummy_bot, interaction):
    ctx = AppCommandContext(dummy_bot, interaction)
    button = Button(style=ButtonStyle.PRIMARY, label="Click", custom_id="a")
    row = ActionRow([button])
    await ctx.send(
        content="hi",
        tts=True,
        components=[row],
        allowed_mentions={"parse": []},
        files=[{"id": 1, "filename": "f.txt"}],
        ephemeral=True,
    )
    dummy_bot._http.create_interaction_response.assert_called_once()
    payload = dummy_bot._http.create_interaction_response.call_args.kwargs[
        "payload"
    ].data.to_dict()
    assert payload["tts"] is True
    assert payload["components"]
    assert payload["allowed_mentions"] == {"parse": []}
    assert payload["attachments"] == [{"id": 1, "filename": "f.txt"}]
    assert payload["flags"] == MessageFlags.EPHEMERAL.value
    await ctx.send_followup(content="again")
    dummy_bot._http.create_followup_message.assert_called_once()


@pytest.mark.asyncio
async def test_second_send_is_followup(dummy_bot, interaction):
    ctx = AppCommandContext(dummy_bot, interaction)
    await ctx.send(content="first")
    await ctx.send_followup(content="second")
    assert dummy_bot._http.create_interaction_response.call_count == 1
    assert dummy_bot._http.create_followup_message.call_count == 1


@pytest.mark.asyncio
async def test_edit_with_components_and_attachments(dummy_bot, interaction):
    ctx = AppCommandContext(dummy_bot, interaction)
    await ctx.send(content="orig")
    row = ActionRow([Button(style=ButtonStyle.PRIMARY, label="B", custom_id="b")])
    await ctx.edit(content="new", components=[row], attachments=[{"id": 1}])
    dummy_bot._http.edit_original_interaction_response.assert_called_once()
    payload = dummy_bot._http.edit_original_interaction_response.call_args.kwargs[
        "payload"
    ]
    assert payload["components"]
    assert payload["attachments"] == [{"id": 1}]


@pytest.mark.asyncio
async def test_send_with_flags(dummy_bot, interaction):
    ctx = AppCommandContext(dummy_bot, interaction)
    await ctx.send(content="hi", flags=MessageFlags.IS_COMPONENTS_V2.value)
    payload = dummy_bot._http.create_interaction_response.call_args.kwargs[
        "payload"
    ].data.to_dict()
    assert payload["flags"] == MessageFlags.IS_COMPONENTS_V2.value


@pytest.mark.asyncio
async def test_send_container_component(dummy_bot, interaction):
    ctx = AppCommandContext(dummy_bot, interaction)
    container = Container(components=[TextDisplay(content="hi")])
    await ctx.send(components=[container], flags=MessageFlags.IS_COMPONENTS_V2.value)
    payload = dummy_bot._http.create_interaction_response.call_args.kwargs[
        "payload"
    ].data.to_dict()
    assert payload["components"][0]["type"] == ComponentType.CONTAINER.value
    assert payload["flags"] == MessageFlags.IS_COMPONENTS_V2.value


@pytest.mark.asyncio
async def test_command_context_edit(command_bot, message):
    async def dummy(ctx):
        pass

    cmd = Command(dummy)
    ctx = CommandContext(
        message=message, bot=command_bot, prefix="!", command=cmd, invoked_with="dummy"
    )
    await ctx.edit(message.id, content="new")
    command_bot._http.edit_message.assert_called_once()
    args = command_bot._http.edit_message.call_args[0]
    assert args[0] == message.channel_id
    assert args[1] == message.id
    assert args[2]["content"] == "new"


@pytest.mark.asyncio
async def test_send_http_error_propagates(dummy_bot, interaction):
    ctx = AppCommandContext(dummy_bot, interaction)
    dummy_bot._http.create_interaction_response.side_effect = RuntimeError("boom")
    with pytest.raises(RuntimeError):
        await ctx.send(content="hi")


@pytest.mark.asyncio
async def test_concurrent_send_only_initial_once(dummy_bot, interaction):
    ctx = AppCommandContext(dummy_bot, interaction)

    async def send_msg(i: int):
        if i == 0:
            await ctx.send(content=str(i))
        else:
            await ctx.send_followup(content=str(i))

    await asyncio.gather(*(send_msg(i) for i in range(50)))
    assert dummy_bot._http.create_interaction_response.call_count == 1
    assert dummy_bot._http.create_followup_message.call_count == 49


@pytest.mark.asyncio
async def test_send_with_flags_2():
    bot = DummyBot()
    interaction = Interaction(
        {
            "id": "1",
            "application_id": bot.application_id,
            "type": InteractionType.APPLICATION_COMMAND.value,
            "token": "tok",
            "version": 1,
        },
        client_instance=bot,
    )
    ctx = AppCommandContext(bot, interaction)
    await ctx.send(content="hi", flags=MessageFlags.IS_COMPONENTS_V2.value)
    payload = bot._http.create_interaction_response.call_args[1][
        "payload"
    ].data.to_dict()
    assert payload["flags"] == MessageFlags.IS_COMPONENTS_V2.value


@pytest.mark.asyncio
async def test_send_container_component_2():
    bot = DummyBot()
    interaction = Interaction(
        {
            "id": "1",
            "application_id": bot.application_id,
            "type": InteractionType.APPLICATION_COMMAND.value,
            "token": "tok",
            "version": 1,
        },
        client_instance=bot,
    )
    ctx = AppCommandContext(bot, interaction)
    container = Container(components=[TextDisplay(content="hi")])
    await ctx.send(
        components=[container],
        flags=MessageFlags.IS_COMPONENTS_V2.value,
    )
    payload = bot._http.create_interaction_response.call_args[1][
        "payload"
    ].data.to_dict()
    assert payload["components"][0]["type"] == ComponentType.CONTAINER.value
    assert payload["flags"] == MessageFlags.IS_COMPONENTS_V2.value



import pytest
import aiohttp
from unittest.mock import AsyncMock

from disagreement.http import HTTPClient
from disagreement import oauth


@pytest.mark.asyncio
async def test_build_authorization_url():
    url = oauth.build_authorization_url(
        client_id="123",
        redirect_uri="https://example.com/cb",
        scope=["identify", "guilds"],
        state="xyz",
    )
    assert url.startswith("https://discord.com/oauth2/authorize?")
    assert "client_id=123" in url
    assert "state=xyz" in url
    assert "scope=identify+guilds" in url


@pytest.mark.asyncio
@pytest.mark.asyncio
async def test_exchange_code_for_token_makes_correct_request(monkeypatch):
    mock_client_response = AsyncMock()
    mock_client_response.status = 200
    mock_client_response.json = AsyncMock(return_value={"access_token": "a"})
    mock_client_response.__aenter__ = AsyncMock(return_value=mock_client_response)
    mock_client_response.__aexit__ = AsyncMock(return_value=None)

    post_mock = AsyncMock(return_value=mock_client_response)
    monkeypatch.setattr("aiohttp.ClientSession.post", post_mock)

    data = await oauth.exchange_code_for_token(
        client_id="id",
        client_secret="secret",
        code="code",
        redirect_uri="https://cb",
    )

    assert data == {"access_token": "a"}
    post_mock.assert_called_once()
    args, kwargs = post_mock.call_args
    assert args[0] == "https://discord.com/api/v10/oauth2/token"
    assert kwargs["headers"]["Content-Type"] == "application/x-www-form-urlencoded"
    assert kwargs["data"]["grant_type"] == "authorization_code"
    assert kwargs["data"]["client_id"] == "id"


@pytest.mark.asyncio
async def test_exchange_code_for_token_custom_session():
    mock_client_response = AsyncMock()
    mock_client_response.status = 200
    mock_client_response.json = AsyncMock(return_value={"access_token": "x"})
    mock_client_response.__aenter__ = AsyncMock(return_value=mock_client_response)
    mock_client_response.__aexit__ = AsyncMock(return_value=None)

    mock_session = AsyncMock()
    mock_session.post = AsyncMock(return_value=mock_client_response)

    data = await oauth.exchange_code_for_token(
        client_id="c1",
        client_secret="c2",
        code="code",
        redirect_uri="https://cb",
        session=mock_session,
    )
    assert data == {"access_token": "x"}
    mock_session.post.assert_called_once()


@pytest.mark.asyncio
async def test_refresh_access_token_success(monkeypatch):
    mock_client_response = AsyncMock()
    mock_client_response.status = 200
    mock_client_response.json = AsyncMock(return_value={"access_token": "b"})
    mock_client_response.__aenter__ = AsyncMock(return_value=mock_client_response)
    mock_client_response.__aexit__ = AsyncMock(return_value=None)

    post_mock = AsyncMock(return_value=mock_client_response)
    monkeypatch.setattr("aiohttp.ClientSession.post", post_mock)

    data = await oauth.refresh_access_token(
        refresh_token="rt",
        client_id="cid",
        client_secret="sec",
    )

    assert data == {"access_token": "b"}
    post_mock.assert_called_once()
    args, kwargs = post_mock.call_args
    assert args[0] == "https://discord.com/api/v10/oauth2/token"
    assert kwargs["data"]["grant_type"] == "refresh_token"
    assert kwargs["data"]["refresh_token"] == "rt"


@pytest.mark.asyncio
async def test_refresh_access_token_error(monkeypatch):
    mock_client_response = AsyncMock()
    mock_client_response.status = 400
    mock_client_response.json = AsyncMock(return_value={"error": "invalid"})
    mock_client_response.__aenter__ = AsyncMock(return_value=mock_client_response)
    mock_client_response.__aexit__ = AsyncMock(return_value=None)

    post_mock = AsyncMock(return_value=mock_client_response)
    monkeypatch.setattr("aiohttp.ClientSession.post", post_mock)

    with pytest.raises(oauth.HTTPException):
        await oauth.refresh_access_token(
            refresh_token="bad",
            client_id="cid",
            client_secret="sec",
        )
